/*
 *   Low-level routines for Wolfson WM8776 codec
 *
 *	Copyright (c) 2007 Rainer Zimmermann <mail@lightshed.de>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 */

/*
 * not implemented:
 * ADC highpass,
 * AGC/limiter functions,
 * DAC output phase+output control
 * de-emphasis mode
 * etc...
 */

#include <asm/io.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/init.h>
#include <sound/core.h>
#include <sound/control.h>
#include <sound/tlv.h>
#include <sound/wm8776.h>

MODULE_AUTHOR("Rainer Zimmermann <mail@lightshed.de>");
MODULE_DESCRIPTION("Routines for control of WM8776 codecs");
MODULE_LICENSE("GPL");

/*
 * If defined, ADC attenuation down to -103dB will be allowed, using the
 * WM8776's digital attenuator. IMO this is rather useless because it will
 * not prevent ADC overdrive, and digital attenuation can always be done in
 *  software.
 * If not defined, ADC level runs from -21 to +24dB, using the analogue PGA.
 *
 * NOTE: this driver does not currently support input mute. Since input level
 * applies also to the analogue bypass besides the ADC, muting should be
 * done using MUTERA/MUTELA in R21, rather than the digital mute using
 * R14/R15.
 */
#undef WM8776_ALLOW_ADC_DIGITAL_ATTEN

/*
 * dB tables
 */
/* headphone output: mute, -73..+6db (1db step) */
static const DECLARE_TLV_DB_SCALE(tlv_scale_headphone, -7400, 100, 1);
/* DAC output: mute, -127..0db (0.5db step) */
static const DECLARE_TLV_DB_SCALE(tlv_scale_dac, -12750, 50, 1);

#ifdef WM8776_ALLOW_ADC_DIGITAL_ATTEN
#define ADC_REG_LOW 0x01
/* ADC gain: mute, -103..+24db (0.5db step) */
static const DECLARE_TLV_DB_SCALE(tlv_scale_adcgain, -10300, 50, 0);
#else
#define ADC_REG_LOW 0xa5
/* ADC gain: mute, -21..+24db (0.5db step) */
static const DECLARE_TLV_DB_SCALE(tlv_scale_adcgain, -2100, 50, 0);
#endif

static struct snd_wm8776_volume_info volumeinfo[WM8776_NUM_VOLUMES] =
{
	[WM8776_VOL_HEADPHONE_L] = {
		.nlvl = 80,
		.tlv_scale = tlv_scale_headphone,
		.reg_idx = WM8776_REG_HEADPHONE_L,
		.reg_mask = 0x7f,
		.reg_low = 0x30,
		.reg_mute = 0x00
	},
	[WM8776_VOL_HEADPHONE_R] = {
		.nlvl = 80,
		.tlv_scale = tlv_scale_headphone,
		.reg_idx = WM8776_REG_HEADPHONE_R,
		.reg_mask = 0x7f,
		.reg_low = 0x30,
		.reg_mute = 0x00
	},
	[WM8776_VOL_DAC_L] = {
		.nlvl = 255,
		.tlv_scale = tlv_scale_dac,
		.reg_idx = WM8776_REG_DAC_ATTEN_L,
		.reg_mask = 0xff,
		.reg_low = 0x01,
		.reg_mute = 0x00
	},
	[WM8776_VOL_DAC_R] = {
		.nlvl = 255,
		.tlv_scale = tlv_scale_dac,
		.reg_idx = WM8776_REG_DAC_ATTEN_R,
		.reg_mask = 0xff,
		.reg_low = 0x01,
		.reg_mute = 0x00
	},
	[WM8776_VOL_ADC_GAIN_L] = {
		.nlvl = 91,
		.tlv_scale = tlv_scale_adcgain,
		.reg_idx = WM8776_REG_ADC_ATTEN_L,
		.reg_mask = 0xff,
		.reg_low = ADC_REG_LOW,
		.reg_mute = ADC_REG_LOW
	},
	[WM8776_VOL_ADC_GAIN_R] = {
		.nlvl = 91,
		.tlv_scale = tlv_scale_adcgain,
		.reg_idx = WM8776_REG_ADC_ATTEN_R,
		.reg_mask = 0xff,
		.reg_low = ADC_REG_LOW,
		.reg_mute = ADC_REG_LOW
	}
};

static struct snd_wm8776_switch_info switchinfo[WM8776_NUM_SWITCHES] =
{
	[WM8776_SW_OUT_DAC] = {
		.n_states = 2,
		.reg_idx = 0x16,
		.reg_inv = 0,
		.reg_mask = 0x01
	},
	[WM8776_SW_OUT_AUX] = {
		.n_states = 2,
		.reg_idx = 0x16,
		.reg_inv = 0,
		.reg_mask = 0x02
	},
	[WM8776_SW_OUT_BYPASS] = {
		.n_states = 2,
		.reg_idx = 0x16,
		.reg_inv = 0,
		.reg_mask = 0x04
	},
	[WM8776_SW_IN_1] = {
		.n_states = 2,
		.reg_idx = 0x15,
		.reg_inv = 0,
		.reg_mask = 0x01
	},
	[WM8776_SW_IN_2] = {
		.n_states = 2,
		.reg_idx = 0x15,
		.reg_inv = 0,
		.reg_mask = 0x02
	},
	[WM8776_SW_IN_3] = {
		.n_states = 2,
		.reg_idx = 0x15,
		.reg_inv = 0,
		.reg_mask = 0x04
	},
	[WM8776_SW_IN_4] = {
		.n_states = 2,
		.reg_idx = 0x15,
		.reg_inv = 0,
		.reg_mask = 0x08
	},
	[WM8776_SW_IN_5] = {
		.n_states = 2,
		.reg_idx = 0x15,
		.reg_inv = 0,
		.reg_mask = 0x10
	}
};

/* write the given register and save the data to the cache */
void snd_wm8776_reg_write(struct snd_wm8776 *wm, unsigned char reg,
						unsigned short val)
{
/*	printk(KERN_DEBUG "wm8776: wm8776_reg_write\n"); */

	mutex_lock(&wm->wm_mutex);
	wm->ops.write(wm->private_data, wm->chip, reg, val);
	wm->regs[reg] = val;
	mutex_unlock(&wm->wm_mutex);
}
EXPORT_SYMBOL(snd_wm8776_reg_write);

/*
 * update the given register with and/or mask and save the data to the cache
 */
void snd_wm8776_reg_mask(struct snd_wm8776 *wm, unsigned char reg,
			unsigned short and_mask, unsigned short or_mask)
{
	unsigned short val = (wm->regs[reg] & and_mask) | or_mask;
	if (val != wm->regs[reg]) {
		mutex_lock(&wm->wm_mutex);
		wm->ops.write(wm->private_data, wm->chip, reg, val);
		wm->regs[reg] = val;
		mutex_unlock(&wm->wm_mutex);
	}
}

/*
 * change a volume setting on a WM8776 codec.
 * index is one of the WM8776_VOL_XXX constants defined in wm8776.h.
 * value runs from 1 to nlvl specified in volume_info, or 0 for mute.
 * returns 1 if changed, 0 otherwise
 */
int snd_wm8776_set_volume(struct snd_wm8776 *wm, int index, int value)
{
	struct snd_wm8776_volume_info *vi = &volumeinfo[index];
	unsigned short flags;

	if (wm->volumes[index] == value)
		return 0;

	wm->volumes[index] = value;

	/*
	 * HEADPHONE (index<2): UPDATE and zero cross enable (0x180),
	 * ADC GAIN: zero cross enable (0x100),
	 * DAC ATTEN: UPDATE (0x100)
	 */
	flags = (index < 2) ? 0x180 : 0x100;
	snd_wm8776_reg_write(wm, vi->reg_idx,
		((value == 0) ? vi->reg_mute : vi->reg_low+value-1) | flags);

	/* handle ADC mute */
	if (index == WM8776_VOL_ADC_GAIN_L)
		snd_wm8776_reg_mask(wm, WM8776_REG_ADC_MUX, ~0x80,
						(value == 0) ? 0x80 : 0);

	if (index == WM8776_VOL_ADC_GAIN_R)
		snd_wm8776_reg_mask(wm, WM8776_REG_ADC_MUX, ~0x40,
						(value == 0) ? 0x40 : 0);

	return 1;
}
EXPORT_SYMBOL(snd_wm8776_set_volume);

int snd_wm8776_get_volume(struct snd_wm8776 *wm, int index)
{
	return wm->volumes[index];
}
EXPORT_SYMBOL(snd_wm8776_get_volume);

const struct snd_wm8776_volume_info *snd_wm8776_get_volume_info(int index)
{
	return volumeinfo+index;
}
EXPORT_SYMBOL(snd_wm8776_get_volume_info);

/* return 1 if changed, 0 otherwise */
int snd_wm8776_set_switch(struct snd_wm8776 *wm, int index, int value)
{
	struct snd_wm8776_switch_info *vi = &switchinfo[index];
	unsigned int mask = 1<<index;
	unsigned int sw = (wm->switches & ~mask) | (value?mask:0);
	if (sw == wm->switches)
		return 0;
	wm->switches = sw;
	if (vi->reg_inv)
		value = !value;
	snd_wm8776_reg_mask(wm, vi->reg_idx, ~vi->reg_mask,
						value?vi->reg_mask:0);
	return 1;
}
EXPORT_SYMBOL(snd_wm8776_set_switch);

int snd_wm8776_get_switch(struct snd_wm8776 *wm, int index)
{
	return (wm->switches & (1<<index)) != 0;
}
EXPORT_SYMBOL(snd_wm8776_get_switch);

const struct snd_wm8776_switch_info *snd_wm8776_get_switch_info(int index)
{
	return switchinfo+index;
}
EXPORT_SYMBOL(snd_wm8776_get_switch_info);

/*
 * this currently sets the same rate for ADC and DAC, but limits ADC rate
 * to 256X (96kHz). For 256X mode (96kHz), this sets ADC oversampling to 64x,
 * as recommended by WM8776 datasheet. Setting the rate is not really
 * necessary in slave mode.
 */
void snd_wm8776_set_rate(struct snd_wm8776 *wm, unsigned char clock_ratio)
{
	unsigned char adc_ratio = clock_ratio;
	if (adc_ratio < WM8776_CLOCK_RATIO_256FS)
		adc_ratio = WM8776_CLOCK_RATIO_256FS;

	snd_wm8776_reg_mask(wm, WM8776_REG_MASTER_MODE_CONTROL,
		0x180,
		adc_ratio | ((adc_ratio == WM8776_CLOCK_RATIO_256FS)?8:0)
				| (clock_ratio<<4));
}
EXPORT_SYMBOL(snd_wm8776_set_rate);


/*
 * reset a WM8776 codec
 * @state: 1 = reset codec only, 0 = reset and restore the registers
 *
 * assert the reset operation and restores the register values to the chips.
 */
void snd_wm8776_reset(struct snd_wm8776 *wm, int state)
{
	unsigned char reg;

	snd_wm8776_reg_write(wm, WM8776_REG_RESET, 0);  /* reset */

	if (state) {
		/* mute DAC */
		snd_wm8776_reg_write(wm, WM8776_REG_DAC_MUTE, 1);
	} else {
		/* update registers from stored values */

		/* restore attenuation levels, with update bit (0x100) set */
		for (reg = 0; reg < 6; reg++)
			snd_wm8776_reg_mask(wm, reg, 0xff, 0x100);

		/* restore all other registers except reset register (0x17) */
		for (reg = 6; reg < 23; reg++)
			snd_wm8776_reg_mask(wm, reg, 0x1ff, 0);
	}
}
EXPORT_SYMBOL(snd_wm8776_reset);

/*
 * initialize a wm8776 chip
 */
void snd_wm8776_init(struct snd_wm8776 *wm, struct snd_card *card,
				void *private_data, int chip,
				const struct snd_wm8776_ops *ops)
{
	static const unsigned short inits_wm8776[] = {
		0x02, 0x100, /* R2: headphone L+R muted + update */
		0x05, 0x100, /* R5: DAC output L+R muted + update */
		0x06, 0x000, /* R6: DAC output phase normal */
		0x07, 0x091, /* R7: DAC enable zero cross detection,
				normal output */
		0x08, 0x000, /* R8: DAC soft mute off */
		0x09, 0x000, /* R9: no deemph, DAC zero detect disabled */
		0x0a, 0x022, /* R10: DAC I2C mode, std polarities, 24bit */
		0x0b, 0x022, /* R11: ADC I2C mode, std polarities, 24bit,
				highpass filter enabled */
		0x0c, 0x042, /* R12: ADC+DAC slave, ADC+DAC 44,1kHz */
		0x0d, 0x000, /* R13: all power up */
		0x0e, 0x100, /* R14: ADC left muted,
				enable zero cross detection */
		0x0f, 0x100, /* R15: ADC right muted,
				enable zero cross detection */
			     /* R16: ALC...*/
		0x11, 0x000, /* R17: disable ALC */
			     /* R18: ALC...*/
			     /* R19: noise gate...*/
		0x15, 0x000, /* R21: ADC input mux init, mute all inputs */
		0x16, 0x001, /* R22: output mux, select DAC */
		0xff, 0xff
	};

	const unsigned short *ptr;
	unsigned char reg;
	unsigned short data;

	mutex_init(&wm->wm_mutex);
	wm->card = card;
	wm->chip = chip;
	wm->private_data = private_data;
	memcpy(&wm->ops, ops, sizeof(wm->ops));

	memset(wm->regs, 0, sizeof(wm->regs));
	memset(wm->volumes, 0, sizeof(wm->volumes));

	/* enable DAC output; mute bypass, aux & all inputs */
	wm->switches = (1<<WM8776_SW_OUT_DAC);

	ptr = inits_wm8776;
	while (*ptr != 0xff) {
		reg = *ptr++;
		data = *ptr++;
		snd_wm8776_reg_write(wm, reg, data);
	}
	printk(KERN_DEBUG "wm8776: init done, wm=%p, priv=%p, chip %i\n",
				wm, wm->private_data, chip);
}
EXPORT_SYMBOL(snd_wm8776_init);


static int __init alsa_wm8776_module_init(void)
{
	return 0;
}

static void __exit alsa_wm8776_module_exit(void)
{
}


module_init(alsa_wm8776_module_init)
module_exit(alsa_wm8776_module_exit)


