/*
 *   ALSA driver for ICEnsemble VT1724 (Envy24HT)
 *
 *   Lowlevel functions for ESI Maya44 cards
 *
 *	Copyright (c) 2007 Rainer Zimmermann <mail@lightshed.de>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 */

#include <asm/io.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <sound/core.h>
#include <sound/wm8776.h>

#include "ice1712.h"
#include "envy24ht.h"
#include "maya44.h"

/*
 * If defined, allow separate lists of allowable rates for ADC/DAC.
 * Actually, ADC & DAC rates are linked, so setting invalid
 * ADC rates (>96 kHz) cannot be prevented in all cases.
 * NOTE: this requires changes in ice1712.h (see ice1712.h.patch)
 * and enabling the corresponding code in ice1724.c
 */
#undef SEPARATE_ADC_RATES

struct maya44_spec {
	struct snd_wm8776 *wm[2];
	unsigned int miodio:1;
};

/*
 * chip addresses on I2C bus
 */
#define WM8776_0_ADDR		0x34		/* Codec 0 */
#define WM8776_1_ADDR		0x36		/* Codec 1 */

/*
 * GPIO pins (known ones for maya44)
 */
#define GPIO_PHANTOM_OFF	(1<<2)
#define GPIO_MIC_RELAY		(1<<4)
#define GPIO_SPDIF_IN_INV	(1<<5)  /* CHECK! */
#define GPIO_MUST_BE_0		(1<<7)


/*
 * WM8776 codecs section
 */

static void maya44_wm_reg_write(void *private_data, int chip,
			   unsigned char addr, unsigned short data)
{
	struct snd_ice1712 *ice = (struct snd_ice1712 *)private_data;

#if 0
	snd_printd(KERN_DEBUG
		   "maya44: maya44_wm_reg_write, ice=%p, chip=%i, addr=%i, "
		   "data=%i\n",
		   ice, chip, (int)addr, (int)data);
#endif

	if (snd_BUG_ON(chip != 0 && chip != 1))
		return;

/*
 * WM8776 registers are up to 9 bits wide, bit 8 is placed in the LSB
 * of the address field
 */
	snd_vt1724_write_i2c(ice,
			     chip ? WM8776_1_ADDR : WM8776_0_ADDR,
			     (addr<<1) | ((data>>8) & 1),
			     data&0xff
			     );
}

/*
 * change the rate on the WM8776 codecs.
 * this assumes that the VT17xx's rate is changed by the calling function.
 * NOTE: even though the WM8776's are running in slave mode and rate
 * selection is automatic, we need to call snd_wm8776_set_rate() here
 * to make sure some flags are set correctly.
 */
static void set_rate(struct snd_ice1712 *ice, unsigned int rate)
{
	struct maya44_spec *maya44_spec = ice->spec;
	int ratio, i;

	switch (rate) {
	case 192000:
		ratio = WM8776_CLOCK_RATIO_128FS;
		break;
	case 176400:
		ratio = WM8776_CLOCK_RATIO_128FS;
		break;
	case 96000:
		ratio = WM8776_CLOCK_RATIO_256FS;
		break;
	case 88200:
		ratio = WM8776_CLOCK_RATIO_384FS;
		break;
	case 48000:
		ratio = WM8776_CLOCK_RATIO_512FS;
		break;
	case 44100:
		ratio = WM8776_CLOCK_RATIO_512FS;
		break;
	case 32000:
		ratio = WM8776_CLOCK_RATIO_768FS;
		break;
	case 0:
		/* no hint - S/PDIF input is master, simply return */
		return;
	default:
		snd_BUG();
		return;
	}

	snd_printd(KERN_DEBUG "maya44: set rate %i, ratio=%d\n", rate, ratio);

	for (i = 0; i < 2; i++)
		snd_wm8776_set_rate(maya44_spec->wm[i], ratio);
}

static int maya44_wm_dev_free(struct snd_device *device)
{
	snd_printd(KERN_DEBUG
		   "maya44: maya44_wm_dev_free, dev=%p, device_data=%p\n",
		   device, device->device_data);
	kfree(device->device_data);
	return 0;
}

static struct wm_vol_control {
	char *name;
	int index[2];
} wm_vol_controls[] =
{
	{ .name = "Crossmix Playback Volume",
	  .index = { WM8776_VOL_HEADPHONE_L, WM8776_VOL_HEADPHONE_R } },
	{ .name = "PCM Playback Volume",
	  .index = { WM8776_VOL_DAC_L, WM8776_VOL_DAC_R } },
	{ .name = "Line Capture Volume",
	  .index = { WM8776_VOL_ADC_GAIN_L, WM8776_VOL_ADC_GAIN_R } },
	{} /* terminator */
};

/* definition of "1 per codec chip" switches */

static struct wm_switch_control {
	char *name;
	int index;
} wm_switch_controls[] =
{
	{ .name = "PCM Playback Switch",	.index = WM8776_SW_OUT_DAC },
/* Aux is not used & set to OFF in wm8776.c by default */
/*	{ .name = "Aux Playback Switch",	.index = WM8776_SW_OUT_AUX },*/
	{ .name = "Bypass Playback Switch", 	.index = WM8776_SW_OUT_BYPASS },
#ifdef _TESTMODE_
	{ .name = "Line Capture Switch 1",  	.index = WM8776_SW_IN_1 },
	{ .name = "Line Capture Switch 2",  	.index = WM8776_SW_IN_2 },
	{ .name = "Line Capture Switch 3",  	.index = WM8776_SW_IN_3	},
	{ .name = "Line Capture Switch 4",  	.index = WM8776_SW_IN_4	},
	{ .name = "Line Capture Switch 5",  	.index = WM8776_SW_IN_5	},
#endif
	{} /* terminator */
};

#define WM_MAKE_PRIV(chip, idx) (((chip)<<8)|(idx))
#define WM_GET_CHIP(priv) ((priv)>>8)
#define WM_GET_IDX(priv) ((priv)&0xff)

#define TLV_HAS_MUTE(tlv) (((tlv)[3]&0x10000) != 0)

static int wm_vol_info(struct snd_kcontrol *kcontrol,
				struct snd_ctl_elem_info *uinfo)
{
	const struct snd_wm8776_volume_info *vi =
		snd_wm8776_get_volume_info(WM_GET_IDX(kcontrol->private_value));

#if 0
	snd_printd(KERN_DEBUG "maya44: wm_vol_info, prv=%i\n",
		   (int)kcontrol->private_value);
#endif

	uinfo->type = SNDRV_CTL_ELEM_TYPE_INTEGER;
	uinfo->count = 2; /* 2 channels per codec */
	uinfo->value.integer.min = TLV_HAS_MUTE(vi->tlv_scale)?0:1;
	uinfo->value.integer.max = vi->nlvl;
	return 0;
}

static int wm_vol_get(struct snd_kcontrol *kcontrol,
				struct snd_ctl_elem_value *ucontrol)
{
	struct snd_ice1712 *ice = snd_kcontrol_chip(kcontrol);
	struct maya44_spec *maya44_spec = ice->spec;
	struct snd_wm8776 *wm;

#if 0
	snd_printd(KERN_DEBUG "maya44: wm_vol_get, ice=%p, prv=%i\n",
		   ice, (int)kcontrol->private_value);
#endif

	/* get L+R volumes from one WM8776 codec */
	wm = maya44_spec->wm[WM_GET_CHIP(kcontrol->private_value)];
	/* left channel */
	ucontrol->value.integer.value[0] =
		snd_wm8776_get_volume(wm,
				      WM_GET_IDX(kcontrol->private_value));
	/* right channel */
	ucontrol->value.integer.value[1] =
		snd_wm8776_get_volume(wm,
				      WM_GET_IDX(kcontrol->private_value)+1);

	return 0;
}

static int wm_vol_put(struct snd_kcontrol *kcontrol,
			struct snd_ctl_elem_value *ucontrol)
{
	struct snd_ice1712 *ice = snd_kcontrol_chip(kcontrol);
	struct maya44_spec *maya44_spec = ice->spec;
	struct snd_wm8776 *wm;
	int changed = 0;

#if 0
	snd_printd(KERN_DEBUG "maya44: wm_vol_put, ice=%p, prv=%i, val=%i,%i\n",
		   ice, (int)kcontrol->private_value,
		   (int)ucontrol->value.integer.value[0],
		   (int)ucontrol->value.integer.value[1]);
#endif

	/* put L+R volumes for one WM8776 codec */
	wm = maya44_spec->wm[WM_GET_CHIP(kcontrol->private_value)];
	/* left channel */
	changed = snd_wm8776_set_volume(wm,
					WM_GET_IDX(kcontrol->private_value),
					ucontrol->value.integer.value[0]);
	/* right channel */
	changed |= snd_wm8776_set_volume(wm,
					 WM_GET_IDX(kcontrol->private_value)+1,
					 ucontrol->value.integer.value[1]);

	return changed;
}

#if 0
static int wm_switch_info(struct snd_kcontrol *kcontrol,
			struct snd_ctl_elem_info *uinfo)
{
	uinfo->type = SNDRV_CTL_ELEM_TYPE_BOOLEAN;
	uinfo->count = 2; /* switches are always one per codec */
	uinfo->value.integer.min = 0;
	uinfo->value.integer.max = 1;
	return 0;
}
#endif

static int wm_switch_get(struct snd_kcontrol *kcontrol,
				struct snd_ctl_elem_value *ucontrol)
{
	struct snd_ice1712 *ice = snd_kcontrol_chip(kcontrol);
	struct maya44_spec *maya44_spec = ice->spec;

#if 0
	snd_printd(KERN_DEBUG "maya44: wm_switch_get, ice=%p, prv=%i\n",
		   ice, (int)kcontrol->private_value);
#endif

	/* get switch from one WM8776 codec */
	ucontrol->value.integer.value[0] =
		snd_wm8776_get_switch(
			maya44_spec->wm[WM_GET_CHIP(kcontrol->private_value)],
			WM_GET_IDX(kcontrol->private_value));

	return 0;
}

static int wm_switch_put(struct snd_kcontrol *kcontrol,
				struct snd_ctl_elem_value *ucontrol)
{
	struct snd_ice1712 *ice = snd_kcontrol_chip(kcontrol);
	struct maya44_spec *maya44_spec = ice->spec;
	int changed = 0;

#if 0
	snd_printd(KERN_DEBUG "maya44: wm_switch_put, ice=%p, prv=%i\n",
		   ice, (int)kcontrol->private_value);
#endif

	/* put switch for one WM8776 codec */
	changed = snd_wm8776_set_switch(
			maya44_spec->wm[WM_GET_CHIP(kcontrol->private_value)],
			WM_GET_IDX(kcontrol->private_value),
			ucontrol->value.integer.value[0]);

	return changed;
}

#ifdef _TESTMODE_

static int gpio_switch_get(struct snd_kcontrol *kcontrol,
				struct snd_ctl_elem_value *ucontrol)
{
	struct snd_ice1712 *ice = snd_kcontrol_chip(kcontrol);
	unsigned int data, mask = 1<<kcontrol->private_value;

#if 0
	snd_printd(KERN_DEBUG "maya44: gpio_switch_get, ice=%p, prv=%i\n",
		   ice, (int)kcontrol->private_value);
#endif

	data = ice->gpio.get_data(ice);
	ucontrol->value.integer.value[0] = ((data & mask) != 0);
	return 0;
}

static int gpio_switch_put(struct snd_kcontrol *kcontrol,
				struct snd_ctl_elem_value *ucontrol)
{
	struct snd_ice1712 *ice = snd_kcontrol_chip(kcontrol);
	unsigned int data, mask = 1<<kcontrol->private_value;
	unsigned int bit = ucontrol->value.integer.value[0]?mask:0;
	int changed = 0;

#if 0
	snd_printd(KERN_DEBUG "maya44: gpio_switch_put, ice=%p, prv=%i\n",
		   ice, (int)kcontrol->private_value);
#endif

	mutex_lock(&ice->gpio_mutex);
	data = ice->gpio.get_data(ice);
	if ((data&mask) != bit) {
		ice->gpio.set_data(ice, (data & ~mask)|bit);
		changed = 1;
	}
	mutex_unlock(&ice->gpio_mutex);
	return changed;
}

#endif /* _TESTMODE_ */

enum { SW_MIC = 0, SW_PHANTOM, SW_SPDIF_IN };

static const char *other_switch_names[] = {
	[SW_MIC] = "Mic/Line Input Switch",
	[SW_PHANTOM] = "Mic Phantom+48V Switch",
/*
 * the next one should probably be called "IEC958 Capture Switch", but
 * then alsamixer wouldn't handle it properly.
 */
	[SW_SPDIF_IN] = "S/PDIF Capture Switch",
	NULL /* terminator */
};

static unsigned char sw_other[] = {
	[SW_MIC] = 0,
	[SW_PHANTOM] = 0,
	[SW_SPDIF_IN] = 0
};

static int set_gpio_bits(struct snd_ice1712 *ice, unsigned int mask,
			 unsigned int bits)
{
	unsigned int data;
	unsigned char changed = 0;
	mutex_lock(&ice->gpio_mutex);
	data = ice->gpio.get_data(ice);
	if ((data&mask) != bits) {
		ice->gpio.set_data(ice, (data & ~mask)|bits);
		changed = 1;
	}
	mutex_unlock(&ice->gpio_mutex);
	return changed;
}

static int other_switch_get(struct snd_kcontrol *kcontrol,
				struct snd_ctl_elem_value *ucontrol)
{
#if 0
	snd_printd(KERN_DEBUG "maya44: other_switch_get, prv=%i\n",
		   (int)kcontrol->private_value);
#endif

	ucontrol->value.integer.value[0] =
				sw_other[(int)kcontrol->private_value];
	return 0;
}

static int other_switch_put(struct snd_kcontrol *kcontrol,
				struct snd_ctl_elem_value *ucontrol)
{
	struct snd_ice1712 *ice = snd_kcontrol_chip(kcontrol);
	struct maya44_spec *maya44_spec = ice->spec;
	int j;
	int set = ucontrol->value.integer.value[0];
	int idx = kcontrol->private_value;

#if 0
	snd_printd(KERN_DEBUG "maya44: other_switch_put, ice=%p, prv=%i\n",
		   ice, (int)kcontrol->private_value);
#endif

	if (set == sw_other[idx])
		return 0; /* not changed */

	switch (kcontrol->private_value) {
	case SW_PHANTOM:
		set_gpio_bits(ice, GPIO_PHANTOM_OFF,
					set ? 0 : GPIO_PHANTOM_OFF);
		break;
	case SW_MIC:
		set_gpio_bits(ice, GPIO_MIC_RELAY,
					set ? GPIO_MIC_RELAY : 0);
		/* line input = IN2, mic input = IN4 */
		for (j = WM8776_SW_IN_1; j <= WM8776_SW_IN_5; j++)
			snd_wm8776_set_switch(
				maya44_spec->wm[0],
				j,
				(j == (set?WM8776_SW_IN_4:WM8776_SW_IN_2)));
		break;
	case SW_SPDIF_IN:
		set_gpio_bits(ice, GPIO_SPDIF_IN_INV,
						set ? 0 : GPIO_SPDIF_IN_INV);
		break;
	default:
		snd_BUG();
		return 0;
	}

	sw_other[idx] = set;
	return 1;
}


/*
 * Maya44 routing switch settings have different meanings than the standard
 * ice1724 switches as defined in snd_vt1724_pro_route_info (ice1724.c).
 */
int snd_vt1724_pro_route_maya44_info(struct snd_kcontrol *kcontrol,
				     struct snd_ctl_elem_info *uinfo)
{
	static char *texts[] = {
		"PCM Out", /* 0 */
		"Input 1", "Input 2", "Input 3", "Input 4"
	};

	uinfo->type = SNDRV_CTL_ELEM_TYPE_ENUMERATED;
	uinfo->count = 1;
	uinfo->value.enumerated.items = 5;
	if (uinfo->value.enumerated.item >= uinfo->value.enumerated.items)
		uinfo->value.enumerated.item =
			uinfo->value.enumerated.items - 1;

	strcpy(uinfo->value.enumerated.name,
					texts[uinfo->value.enumerated.item]);
	return 0;
}

int maya44_route_shift(int idx)
{
	static const unsigned char shift[10] =
				{ 8, 20, 0, 3, 11, 23, 14, 26, 17, 29 };
	return shift[idx%10];
}

static int snd_vt1724_pro_route_maya44_get(struct snd_kcontrol *kcontrol,
					   struct snd_ctl_elem_value *ucontrol)
{
	struct snd_ice1712 *ice = snd_kcontrol_chip(kcontrol);
	int idx = snd_ctl_get_ioffidx(kcontrol, &ucontrol->id);
	ucontrol->value.enumerated.item[0] =
		snd_ice1724_get_route_val(ice, maya44_route_shift(idx));
	return 0;
}

static int snd_vt1724_pro_route_maya44_put(struct snd_kcontrol *kcontrol,
					   struct snd_ctl_elem_value *ucontrol)
{
	struct snd_ice1712 *ice = snd_kcontrol_chip(kcontrol);
	int idx = snd_ctl_get_ioffidx(kcontrol, &ucontrol->id);
	return snd_ice1724_put_route_val(ice,
					 ucontrol->value.enumerated.item[0],
					 maya44_route_shift(idx));
}


struct snd_kcontrol_new snd_vt1724_mixer_pro_maya44_route __devinitdata =
{
	.iface = SNDRV_CTL_ELEM_IFACE_MIXER,
	.name = "H/W Playback Route",
	.info = snd_vt1724_pro_route_maya44_info,
	.get = snd_vt1724_pro_route_maya44_get,
	.put = snd_vt1724_pro_route_maya44_put,
#if 1
	.count = 10  /* XXX check whether controls 5-9 have any meaning */
#else
	.count = 4
#endif
};

static int __devinit maya44_add_controls(struct snd_ice1712 *ice)
{
	int err, i;
	struct snd_kcontrol_new newc;
	struct wm_vol_control *pv;
	struct wm_switch_control *ps;

#if 0
	snd_printd(KERN_DEBUG "maya44: maya44_add_controls, ice=%p\n", ice);
#endif

	/* add volume controls for both WM8776 codecs */

	memset(&newc, 0, sizeof(newc));
	newc.iface = SNDRV_CTL_ELEM_IFACE_MIXER,
	newc.access = SNDRV_CTL_ELEM_ACCESS_READWRITE |
					SNDRV_CTL_ELEM_ACCESS_TLV_READ,
	newc.info = wm_vol_info;
	newc.get = wm_vol_get;
	newc.put = wm_vol_put;
	newc.count = 1;

	for (pv = wm_vol_controls; pv->name != NULL; pv++) {
		newc.name = pv->name;
		newc.tlv.p =
			snd_wm8776_get_volume_info(pv->index[0])->tlv_scale;

		for (newc.index = 0; newc.index < 2; newc.index++) {
			newc.private_value =
				WM_MAKE_PRIV(newc.index, pv->index[0]);
#if 0
			snd_printd(KERN_DEBUG
				"maya44: add control '%s' idx %i, pval=%i\n",
				pv->name, (int)newc.index,
				(int)newc.private_value);
#endif

			err = snd_ctl_add(ice->card, snd_ctl_new1(&newc, ice));
#if 0
			snd_printd(KERN_DEBUG "maya44: err=%i\n", err);
#endif
			if (err < 0)
				return err;
		}
	}

	/* add switch controls for both WM8776 codecs */

	memset(&newc, 0, sizeof(newc));
	newc.iface = SNDRV_CTL_ELEM_IFACE_MIXER,
	newc.access = SNDRV_CTL_ELEM_ACCESS_READWRITE,
	newc.info = /*wm_switch_info*/ snd_ctl_boolean_mono_info;
	newc.get = wm_switch_get;
	newc.put = wm_switch_put;
	newc.count = 1;

	for (ps = wm_switch_controls; ps->name != NULL; ps++) {
		newc.name = ps->name;
		for (newc.index = 0; newc.index < 2; newc.index++) {
			newc.private_value =
					WM_MAKE_PRIV(newc.index, ps->index);
			err = snd_ctl_add(ice->card, snd_ctl_new1(&newc, ice));
			if (err < 0)
				return err;
		}
	}

#ifdef _TESTMODE_
	/* add GPIO controls */

	memset(&newc, 0, sizeof(newc));
	newc.iface = SNDRV_CTL_ELEM_IFACE_MIXER,
	newc.access = SNDRV_CTL_ELEM_ACCESS_READWRITE,
	newc.info = snd_ctl_boolean_mono_info; /* simple switch */
	newc.get = gpio_switch_get;
	newc.put = gpio_switch_put;
	newc.count = 1;

	{
		char gpname[40];
		int i;
		for (i = 0; i < 24; i++) {
			sprintf(gpname, "GPIO bit %02i", i);
			newc.name = gpname;
			newc.private_value = i;

			err = snd_ctl_add(ice->card, snd_ctl_new1(&newc, ice));
			if (err < 0)
				return err;
		}
	}
#endif

	/* add other controls */

	memset(&newc, 0, sizeof(newc));
	newc.iface = SNDRV_CTL_ELEM_IFACE_MIXER,
	newc.access = SNDRV_CTL_ELEM_ACCESS_READWRITE,
	newc.info = snd_ctl_boolean_mono_info; /* simple switch */
	newc.get = other_switch_get;
	newc.put = other_switch_put;
	newc.count = 1;

	for (i = 0; (newc.name = (char *)other_switch_names[i]) != NULL; i++) {
		newc.private_value = i;
#if 0
		snd_printd(KERN_DEBUG "maya44: add control '%s', i=%i\n",
			   newc.name, i);
#endif
		err = snd_ctl_add(ice->card, snd_ctl_new1(&newc, ice));
		if (err < 0)
			return err;
	}

	return 0;
}

/*
 * initialize the chip
 */
static int __devinit maya44_init(struct snd_ice1712 *ice)
{
	int err, i, j;
	struct maya44_spec *maya44_spec;
	static unsigned int rates[] = {
		32000, 44100, 48000, 64000, 88200, 96000, 176400, 192000
	};
#ifdef SEPARATE_ADC_RATES
	/* record rates: 32..96 kHz */
	static struct snd_pcm_hw_constraint_list adc_rates = {
		.count = ARRAY_SIZE(rates)-2,
		.list = rates,
		.mask = 0
	};
#endif
	/* playback rates: 32..192 kHz */
	static struct snd_pcm_hw_constraint_list dac_rates = {
		.count = ARRAY_SIZE(rates),
		.list = rates,
		.mask = 0
	};

	static struct snd_wm8776_ops _wm_ops = { .write = maya44_wm_reg_write };
	static struct snd_device_ops _dev_ops = {
		.dev_free = maya44_wm_dev_free,
		/* .dev_register=..., */
		/* .dev_disconnect=... */
	};

	snd_printd(KERN_DEBUG "maya44: maya44_init start\n");

	maya44_spec = kzalloc(sizeof(*maya44_spec), GFP_KERNEL);
	if (!maya44_spec)
		return -ENOMEM;
	ice->spec = maya44_spec;

	/* initialise codecs */
	ice->num_total_dacs = 4;
	ice->num_total_adcs = 4;
	ice->akm_codecs = 0;

	for (i = 0; i < 2; i++) {
		struct snd_wm8776 *wm;
		wm = maya44_spec->wm[i] =
			kzalloc(sizeof(struct snd_wm8776), GFP_KERNEL);
		if (!wm)
			return -ENOMEM;
		err = snd_device_new(ice->card, SNDRV_DEV_LOWLEVEL,
							wm, &_dev_ops);
		if (err < 0) {
			kfree(wm);
			return err;
		}
		snd_wm8776_init(wm, ice->card, ice, i, &_wm_ops);
		snd_printd(KERN_DEBUG "maya44: wm8776_init %i done\n", i);

		/* Maya44 line inputs are seen on IN2 */
		for (j = WM8776_SW_IN_1; j <= WM8776_SW_IN_5; j++)
			snd_wm8776_set_switch(wm, j, (j == WM8776_SW_IN_2));
	}

/*
 * set card specific rates.
 */
#ifdef SEPARATE_ADC_RATES
	ice->hw_rates = &dac_rates;
	ice->hw_rates_adc = &adc_rates;
#else
	ice->hw_rates = &dac_rates;
#endif

	/* register change rate notifier */
	ice->gpio.set_pro_rate = set_rate;

	/* RDMA1 (2nd input channel) is used for ADC by default */
	ice->force_rdma1 = 1;

	snd_printd(KERN_DEBUG "maya44: maya44_init finished\n");

	return 0;
}


/*
 * Maya44 boards don't provide the EEPROM data except for the vendor IDs.
 * hence the driver needs to sets up it properly.
 */

static unsigned char maya44_eeprom[] __devinitdata = {
	[ICE_EEP2_SYSCONF]     = 0x45,
		/* clock xin1=49.152MHz, mpu401, 2 stereo ADCs+DACs */
	[ICE_EEP2_ACLINK]      = 0x80,
		/* I2S */
	[ICE_EEP2_I2S]         = 0xf8,
		/* vol, 96k, 24bit, 192k */
	[ICE_EEP2_SPDIF]       = 0xc3,
		/* enable spdif out, spdif out supp, spdif-in, ext spdif out */
	[ICE_EEP2_GPIO_DIR]    = 0xff,
	[ICE_EEP2_GPIO_DIR1]   = 0xff,
	[ICE_EEP2_GPIO_DIR2]   = 0xff,
	[ICE_EEP2_GPIO_MASK]   = 0/*0x9f*/,
	[ICE_EEP2_GPIO_MASK1]  = 0/*0xff*/,
	[ICE_EEP2_GPIO_MASK2]  = 0/*0x7f*/,
	[ICE_EEP2_GPIO_STATE]  = GPIO_PHANTOM_OFF | GPIO_SPDIF_IN_INV,
	[ICE_EEP2_GPIO_STATE1] = 0x00,
	[ICE_EEP2_GPIO_STATE2] = 0x00,
};

/* entry point */
struct snd_ice1712_card_info snd_vt1724_maya44_cards[] __devinitdata = {
	{
		.subvendor = VT1724_SUBDEVICE_MAYA44,
		.name = "ESI Maya44",
		.model = "maya44",
		.chip_init = maya44_init,
		.build_controls = maya44_add_controls,
		.eeprom_size = sizeof(maya44_eeprom),
		.eeprom_data = maya44_eeprom,
	},
	{ } /* terminator */
};
