/*
 *  Creative SoundBlaster X-Fi driver (emu20k1 chipset)
 *
 *  Copyright(c) 2008 Takashi Iwai <tiwai@suse.de>
 *
 *  Based on sbxfi driver of OSS v4.0 OSS version:
 *    Copyright (C) 4Front Technologies 1996-2008.
 *    Copyright (C) Creative Technology, Ltd., 2007. All rights reserved.
 *
 *  This this source file is released under GPL v2 license (no other versions).
 *  See the COPYING file included in the main directory of this source
 *  distribution for the license terms and conditions.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/mutex.h>
#include <linux/dma-mapping.h>
#include <linux/moduleparam.h>
#include <linux/pci.h>
/* #include <linux/math64.h> */
#include <linux/log2.h>
#include <sound/core.h>
#include <sound/pcm.h>
#include <sound/control.h>
#include <sound/initval.h>
#include <sound/info.h>
#include <sound/tlv.h>
#include "emu20k1-regs.h"

/*
 */

MODULE_AUTHOR("Takashi Iwai <tiwai@suse.de>");
MODULE_DESCRIPTION("Creative SoundBlaster X-Fi");
MODULE_LICENSE("GPL");
MODULE_SUPPORTED_DEVICE("{{Creative Labs,SB X-Fi}}");

static int index[SNDRV_CARDS] = SNDRV_DEFAULT_IDX;
static char *id[SNDRV_CARDS] = SNDRV_DEFAULT_STR;
static int enable[SNDRV_CARDS] = SNDRV_DEFAULT_ENABLE_PNP;
static int base_rate = 48000;
static int debug = 1;

module_param_array(index, int, NULL, 0444);
MODULE_PARM_DESC(index, "Index value for SB X-Fi sound driver.");
module_param_array(id, charp, NULL, 0444);
MODULE_PARM_DESC(id, "ID string for SB X-Fi sound driver.");
module_param_array(enable, bool, NULL, 0444);
MODULE_PARM_DESC(enable, "Enable SB X-Fi sound driver.");
module_param(base_rate, int, 0444);
MODULE_PARM_DESC(base_rate, "Base sample rate for SB X-Fi sound driver.");
module_param(debug, int, 0644);
MODULE_PARM_DESC(debug, "Debug flag: 0 = disable");

/*
 * Build options
 */

/* continuous rate support: BROKEN */
#undef XXX_CONT_RATE

/* multi playback - not implemented */
#undef XXX_MULTI_STREAMS

/* use system timer */
#undef XXX_SYSTEM_TIMER

/* constant ticks */
/* #define XXX_CONST_TICKS		(96000 * 5 / 1000) */


/*
 */

#define PFX	"sbxfi: "

enum {
	SBXFI_HW_ORIG,
	SBXFI_HW_055x,
	SBXFI_HW_073x,
	SBXFI_HW_UAA,
	SBXFI_HW_TYPES
};

#define SBXFI_TLB_PAGES		1024
#define SBXFI_PAGE_SIZE		4096
#define SBXFI_MAX_BUFFER	(SBXFI_TLB_PAGES * SBXFI_PAGE_SIZE)
#define SBXFI_MAX_SRCS		128	/* 256 / 2 */
#define SBXFI_TIMER_FREQ	48000
#define SBXFI_DMA_MASK		DMA_32BIT_MASK

#define MAX_CHANNELS		2

/* port state */
enum {
	PORT_OPENED,
	PORT_PREPARED,
	PORT_RUNNING
};

/* ADC source */
enum {
	ADC_SRC_MICIN,
	ADC_SRC_LINEIN,
	ADC_SRC_AUX,
	ADC_SRC_VIDEO,
	ADC_SRC_DIG,
	ADC_SRC_NUMS,
};

struct sbxfi_port {
	struct snd_pcm_substream *substream;
	/* mapping */
	unsigned int src[MAX_CHANNELS];
	unsigned int conv[MAX_CHANNELS];
	unsigned int vol[MAX_CHANNELS];
	/* TLB entry */
	int tlb_index;
	unsigned int tlb_pages;
	/* buffer pointers */
	unsigned int start_addr;
	unsigned int buffer_bytes;
	/* status */
	int state;
	int need_update;
	unsigned int last_pos;
	unsigned int frag_count;
	struct list_head list;
};

enum {
	VOLUME_MASTER,
	VOLUME_PCM,
	VOLUME_CAPTURE,
	VOLUME_NUM_TYPES
};

struct sbxfi {
	struct snd_card *card;
	struct pci_dev *pci;
	int bar;
	long port;
	int irq;
	int type;
	struct snd_pcm *pcm;

	spinlock_t reg_lock;	/* register access lock */
	spinlock_t timer_lock;
	spinlock_t port_lock;
	struct mutex tlb_mutex;

	int irq_handling;
	int timer_running;
	int timer_reprogram;

	unsigned int base_rate;
	int vol[VOLUME_NUM_TYPES][2];
	int capsrc;
	int micboost;

	struct snd_dma_buffer tlb;
	int used_pages;
	DECLARE_BITMAP(tlb_page_bitmap, SBXFI_TLB_PAGES);
	DECLARE_BITMAP(src_bitmap, SBXFI_MAX_SRCS);
	struct page *silent_page;
	u32 silent_addr;	/* 32bit address space */

	/* linked list of ports; protected via port_lock */
	struct list_head port_list;

#ifdef XXX_SYSTEM_TIMER
	struct timer_list systimer;
#endif
};


/*
 * debug: 0 = off, 1 = normal debug, 2 = verbose, 3 = log all reg access
 */
#define LOG(lvl,fmt,args...) do {   \
	if (debug >= (lvl)) \
		printk(KERN_DEBUG "SBXFI: " fmt, ##args); \
	} while (0)


/*
 * Low-level accessors
 */

static inline unsigned int __sbxfi_read(struct sbxfi *chip, unsigned int reg,
					unsigned int ptr_ofs,
					unsigned int data_ofs)
{
	unsigned int val;
	unsigned long flags;

	spin_lock_irqsave(&chip->reg_lock, flags);
	outl(reg, chip->port + ptr_ofs);
	val = inl(chip->port + data_ofs);
	spin_unlock_irqrestore(&chip->reg_lock, flags);
	LOG(3, "  read(%x) 0x%x = 0x%x\n", ptr_ofs, reg, val);
	return val;
}

static unsigned int sbxfi_read_pci(struct sbxfi *chip, unsigned int reg)
{
	return __sbxfi_read(chip, reg, IO_PCI_CONF_ADDR, IO_PCI_CONF_DATA);
}

static unsigned int sbxfi_read(struct sbxfi *chip, unsigned int reg)
{
	return __sbxfi_read(chip, reg, IO_PTR, IO_DATA);
}

static inline void __sbxfi_write(struct sbxfi *chip, unsigned int reg,
				 unsigned int val,
				 unsigned int ptr_ofs, unsigned int data_ofs)
{
	unsigned long flags;
	LOG(3, "  write(%x) 0x%x = 0x%x\n", ptr_ofs, reg, val);
	spin_lock_irqsave(&chip->reg_lock, flags);
	outl(reg, chip->port + ptr_ofs);
	outl(val, chip->port + data_ofs);
	spin_unlock_irqrestore(&chip->reg_lock, flags);
}

static void sbxfi_write_pci(struct sbxfi *chip, unsigned int reg,
			    unsigned int val)
{
	__sbxfi_write(chip, reg, val, IO_PCI_CONF_ADDR, IO_PCI_CONF_DATA);
}

static void sbxfi_write(struct sbxfi *chip, unsigned int reg, unsigned int val)
{
	__sbxfi_write(chip, reg, val, IO_PTR, IO_DATA);
}

/*
 * Timer and interrupt handling
 */

#ifdef XXX_SYSTEM_TIMER

static void sbxfi_handle_timer_callback(struct sbxfi *chip);

static void sbxfi_timer_handle(unsigned long data)
{
	LOG(2, "TIMER ISR\n");
	sbxfi_handle_timer_callback((struct sbxfi *)data);
}

static void sbxfi_init_timer(struct sbxfi *chip)
{
	setup_timer(&chip->systimer, sbxfi_timer_handle, (unsigned long)chip);
}

static void sbxfi_stop_timer(struct sbxfi *chip)
{
	if (chip->timer_running) {
		LOG(2, "STOP TIMER\n");
		del_timer(&chip->systimer);
		chip->timer_running = 0;
	}
}

static void sbxfi_rearm_timer(struct sbxfi *chip, int ticks)
{
	ticks = msecs_to_jiffies((ticks * 1000) / SBXFI_TIMER_FREQ);
	LOG(2, "START TIMER: ticks(HZ) = %d\n", ticks);
	if (!chip->timer_running) {
		chip->systimer.expires = jiffies + ticks;
		add_timer(&chip->systimer);
	} else
		mod_timer(&chip->systimer, jiffies + ticks);
	chip->timer_running = 1;
}

#else

#define MAX_TICKS	((1 << 13) - 1)

static void sbxfi_init_timer(struct sbxfi *chip)
{
}

static void sbxfi_set_timer(struct sbxfi *chip, int ticks)
{
	LOG(2, "SET TIMER TICKS = %d\n", ticks);
	if (ticks > MAX_TICKS)
		ticks = MAX_TICKS;
	sbxfi_write(chip, TIMR, ticks | TIMR_IE | TIMR_IP);
}

static void sbxfi_stop_timer(struct sbxfi *chip)
{
	if (chip->timer_running) {
		LOG(2, "STOP EMU TIMER\n");
		sbxfi_write(chip, TIMR, 0);
		sbxfi_write(chip, GLOBAL_INT_ENABLE, 0);
		chip->timer_running = 0;
	}
}

static void sbxfi_rearm_timer(struct sbxfi *chip, int ticks)
{
	sbxfi_set_timer(chip, ticks);
	if (!chip->timer_running)
		sbxfi_write(chip, GLOBAL_INT_ENABLE, IT_INT);
	chip->timer_running = 1;
}

#endif /* XXX_SYSTEM_TIMER */

static snd_pcm_uframes_t
sbxfi_pcm_pointer(struct snd_pcm_substream *substream);

/*
 * reprogram the timer interval;
 * checks the port list and determines the next timer interval
 * also updates the each stream position, returns the number of streams
 * to call snd_pcm_period_elapsed() appropriately
 *
 * call this inside the timerlock and irq disabled
 */
static int sbxfi_reprogram_timer(struct sbxfi *chip)
{
	struct sbxfi_port *port;
	int min_intr = -1;
	int updates = 0;
	
	spin_lock(&chip->port_lock);
	list_for_each_entry(port, &chip->port_list, list) {
		struct snd_pcm_runtime *runtime;
		unsigned int pos, diff;
		int intr;
		if (port->state != PORT_RUNNING)
			continue;
		runtime = port->substream->runtime;
		pos = sbxfi_pcm_pointer(port->substream);
		if (pos < port->last_pos)
			diff = runtime->buffer_size - port->last_pos + pos;
		else
			diff = pos - port->last_pos;
		port->last_pos = pos;
		while (diff >= port->frag_count) {
			port->frag_count += runtime->period_size;
			port->need_update = 1;
			updates++;
		}
		port->frag_count -= diff;
#ifdef XXX_CONST_TICKS
		intr = XXX_CONST_TICKS;
#else
		/*
		 * if we support generic rate...
		 * intr = div_u64((u64)port->frag_count * SBXFI_TIMER_FREQ,
		 *			runtime->rate);
		*/
		intr = port->frag_count / (runtime->rate / SBXFI_TIMER_FREQ);
#endif
		if (min_intr < 0 || intr < min_intr)
			min_intr = intr;
	}
	spin_unlock(&chip->port_lock);
	
	if (min_intr > 0)
		sbxfi_rearm_timer(chip, min_intr);
	else
		sbxfi_stop_timer(chip);
	chip->timer_reprogram = 0; /* clear flag */
	return updates;
}

/* look through the port list and call period_elapsed if needed */
static void sbxfi_check_period_update(struct sbxfi *chip)
{
	struct sbxfi_port *port;
	unsigned long flags;

	spin_lock_irqsave(&chip->port_lock, flags);
	list_for_each_entry(port, &chip->port_list, list) {
		if (port->need_update) {
			port->need_update = 0;
			if (port->state == PORT_RUNNING)
				snd_pcm_period_elapsed(port->substream);
		}
	}
	spin_unlock_irqrestore(&chip->port_lock, flags);
}

/* Handle timer updates */
static void sbxfi_handle_timer_callback(struct sbxfi *chip)
{
	int update;
	unsigned long flags;

	spin_lock_irqsave(&chip->timer_lock, flags);
	chip->irq_handling = 1;
	do {
		update = sbxfi_reprogram_timer(chip);
		spin_unlock(&chip->timer_lock);
		if (update)
			sbxfi_check_period_update(chip);
		spin_lock(&chip->timer_lock);
	} while (chip->timer_reprogram);
	chip->irq_handling = 0;
	spin_unlock_irqrestore(&chip->timer_lock, flags);
}

/* IRQ handler */
static irqreturn_t sbxfi_interrupt(int irq, void *dev_id)
{
	struct sbxfi *chip = dev_id;
	unsigned int status;

	status = sbxfi_read(chip, GLOBAL_INT_PENDING);
	if (!status)
		return IRQ_NONE;

	LOG(2, "IRQ = 0x%x\n", status);
	/* SRC interrupt(s) pending */
	if (status & SRC_INT) {
		unsigned int map, srcip;
		int i;
		map = sbxfi_read(chip, SRC_INT_PENDING_MAP);
		for (i = 0; i < 7; i++)
			if (map & (1 << i)) {
				/* up to 32 streams */
				srcip = sbxfi_read(chip, SRCIP(i));
				sbxfi_write(chip, SRCIP(i), srcip); /* ack */
			}
	}

	/* Interval timer interrupt */
	if (status & IT_INT)
		sbxfi_handle_timer_callback(chip);

	/* ack */
	sbxfi_write(chip, GLOBAL_INT_PENDING, status);
	return IRQ_HANDLED;
}

/* start/stop the timer */
static void sbxfi_update_timer(struct sbxfi *chip)
{
	unsigned long flags;
	int update;

	if (chip->irq_handling) {
		/* reached from IRQ handler; let it handle later */
		chip->timer_reprogram = 1;
		return;
	}

	spin_lock_irqsave(&chip->timer_lock, flags);
	sbxfi_stop_timer(chip);
	update = sbxfi_reprogram_timer(chip);
	spin_unlock_irqrestore(&chip->timer_lock, flags);
	if (update)
		sbxfi_check_period_update(chip);
}

/*
 */

#define CA_OFFSET	0x80

static void sbxfi_setup_src(struct sbxfi *chip, unsigned int ch,
			    unsigned int start_addr, unsigned int loop_addr,
			    unsigned int cache, unsigned int ctrl)
{
	LOG(1, "SETUP SRC: ch=0x%x, start=0x%x, loop=0x%x, cache=0x%x, ctrl=0x%x\n",
	    ch, start_addr, loop_addr, cache, ctrl);
	sbxfi_write(chip, SRCCA(ch), start_addr + CA_OFFSET); /* current addr */
	sbxfi_write(chip, SRCCF(ch), 0); /* current fraction */
	sbxfi_write(chip, SRCSA(ch), start_addr);
	sbxfi_write(chip, SRCLA(ch), loop_addr);
	sbxfi_write(chip, SRCCCR(ch), cache); /* cache control */
	sbxfi_write(chip, SRCCTL(ch), ctrl); /* SRC control */
}

#define CRM_TIMESLOT_ALLOC_BLOCK_SIZE   16
#define CRM_PTS_PITCH                   6
#define CRM_PARAM_SRC_OFFSET            0x60

static unsigned int sbxfi_get_param_pitch_chan(int id)
{
	int interp = ((id * CRM_TIMESLOT_ALLOC_BLOCK_SIZE) + CRM_PTS_PITCH) -
		CRM_PARAM_SRC_OFFSET;
	if (interp < 0)
		interp += 4096;
	return interp;
}

static unsigned int sbxfi_get_audio_src_chan(int src)
{
	/* SRC channel is in Audio Ring slot 1, after every 16 slot */
	return (src << 4) + SRC_CH_0;
}

static unsigned int sbxfi_get_audio_sum_chan(int chn)
{
	/* SUM channel is in Audio Ring slot 0xc, after every 16 slot */
	return (chn << 4) + MIXER_SUM_CH_0;
}

/* set audio mixer with y-immediate const */
static void sbxfi_write_amop(struct sbxfi *chip,
			     unsigned int xdata, unsigned int ydata,
			     unsigned int chn, unsigned int hidata)
{
	sbxfi_write(chip, AMOP_START(chn),
		    (ydata << 18) | (xdata << 4) | 1);
	sbxfi_write(chip, AMOP_START(chn) + 4, hidata);
}

/*
 */
static void sbxfi_playback_start(struct sbxfi *chip, struct sbxfi_port *port)
{
	unsigned int start, loop;
	unsigned int ctrl, ratec = 0;
	struct snd_pcm_runtime *runtime;

	if (port->tlb_index < 0)
		return;
	runtime = port->substream->runtime;
	start = port->start_addr;
	loop = start + port->buffer_bytes;

	switch (runtime->rate) {
	case 48000:
		ratec = SRCCTL_REF_48K;
		break;
	case 96000:
		ratec = SRCCTL_REF_96K;
		break;
	case 192000:
		ratec = SRCCTL_REF_192K;
		break;
	}
	ratec |= SRCCTL_FMT_S16;
	ratec |= SRCCTL_BUS_MASTER;
	ratec |= SRCCTL_STATE_INIT;

	ctrl = ratec;
	ctrl |= SRCCTL_PITCH_MASTER; /* master */
	ctrl |= (1 << SRCCTL_ILSZ_SHIFT); /* interleaved stereo */
	sbxfi_setup_src(chip, port->src[0], start, loop, 0x100, ctrl);

	/* slave */
	ctrl = ratec;
	sbxfi_setup_src(chip, port->src[1], start, loop, 0x100, ctrl);
}

static void sbxfi_playback_stop(struct sbxfi *chip, struct sbxfi_port *port)
{
	unsigned int val;

	val = sbxfi_read(chip, SRCCTL(port->src[0]));
	val &= ~0x0f;
	val &= ~SRCCTL_IE; /* Interrupt disable */
	sbxfi_write(chip, SRCCTL(port->src[0]), val);

	val = sbxfi_read(chip, SRCCTL(port->src[1]));
	val &= ~0x0f;
	sbxfi_write(chip, SRCCTL(port->src[1]), val);
}


static void sbxfi_capture_start(struct sbxfi *chip, struct sbxfi_port *port)
{
	unsigned int start, loop;
	unsigned int ctrl, ratec = 0;
	struct snd_pcm_runtime *runtime;

	if (port->tlb_index < 0)
		return;
	runtime = port->substream->runtime;
	start = port->start_addr;
	loop = start + port->buffer_bytes;

	switch (runtime->rate) {
	case 48000:
		ratec = SRCCTL_REF_48K;
		break;
	case 96000:
		ratec = SRCCTL_REF_96K;
		break;
	case 192000:
		ratec = SRCCTL_REF_192K;
		break;
	}
	ratec |= SRCCTL_WR;
	ratec |= SRCCTL_FMT_S16;
	ratec |= SRCCTL_BUS_MASTER;
	ratec |= SRCCTL_STATE_RUN;

	ctrl = ratec;
	ctrl |= SRCCTL_PITCH_MASTER; /* master */
	ctrl |= (1 << SRCCTL_ILSZ_SHIFT);	/* Interleaved stereo */
	sbxfi_setup_src(chip, port->src[0], start, loop, 0x100, ctrl);

	/* slave */
	ctrl = ratec;
	sbxfi_setup_src(chip, port->src[1], start, loop, 0x80, ctrl);
#if 0 /* XXX */
	/* Enable SRC input from Audio Ring */
	sbxfi_write(chip, SRC_MASTER_CTL, 0x1);
#endif
}

static void sbxfi_capture_stop(struct sbxfi *chip, struct sbxfi_port *port)
{
	unsigned int val;

	val = sbxfi_read(chip, SRCCTL(port->src[0]));
	val &= ~0x0f;
	sbxfi_write(chip, SRCCTL(port->src[0]), val);

	val = sbxfi_read(chip, SRCCTL(port->src[1]));
	val &= ~0x0f;
	sbxfi_write(chip, SRCCTL(port->src[1]), val);
#if 0 /* XXX */
	/* Disable SRC inputs from Audio Ring */
	sbxfi_write(chip, SRC_MASTER_CTL, 0x0);
#endif
}

/* some sync of reset sequence */
static int sbxfi_sync_i2c(struct sbxfi *chip)
{
	unsigned long timeout = jiffies + msecs_to_jiffies(1000);
	do {
		if (sbxfi_read_pci(chip, 0xec) & 0x800000)
			return 0;
		udelay(10);
		cond_resched();
	} while (time_after(timeout, jiffies));
	printk(KERN_ERR "XFi DAC reset timeout\n");
	return -EIO;
}

static void sbxfi_prepare_init(struct sbxfi *chip)
{
	sbxfi_write_pci(chip, 0xcc, 0x8c);
	sbxfi_write_pci(chip, 0xcc, 0x0e);
	if (((sbxfi_read_pci(chip, 0xcc)) & 0xff) != 0xaa) {
		sbxfi_write_pci(chip, 0xcc, 0xee);
		sbxfi_write_pci(chip, 0xcc, 0xaa);
	}
	if (((sbxfi_read_pci(chip, 0xcc)) & 0xff) != 0xaa)
		return;

	/* write to i2c status control */
	sbxfi_write_pci(chip, 0xec, 0x05);
	sbxfi_sync_i2c(chip);
}

static void sbxfi_set_dac_cmd(struct sbxfi *chip, unsigned int cmd)
{
	sbxfi_write_pci(chip, 0xe0, 0x00180080);
	sbxfi_write_pci(chip, 0xe4, cmd);
	sbxfi_sync_i2c(chip);
}

static void sbxfi_reset_dac(struct sbxfi *chip)
{
	unsigned int gpio;
	int i;

	LOG(1, "RESET DAC\n");
	sbxfi_prepare_init(chip);

	/* To be effective, need to reset the DAC twice -- why? */
	for (i = 0; i < 2; i++) {
		mdelay(100); /* FIXME */
		gpio = sbxfi_read(chip, GPIO);
		gpio &= ~GPIO_O_RESET_DAC;
		sbxfi_write(chip, GPIO, gpio);
		sbxfi_read(chip, GPIO); /* post */
		mdelay(1);
		sbxfi_write(chip, GPIO, gpio | GPIO_O_RESET_DAC);
	}

	sbxfi_set_dac_cmd(chip, 0x8001);
	sbxfi_set_dac_cmd(chip, 0x1002);
}

static void sbxfi_init_dac(struct sbxfi *chip)
{
	unsigned int gpio, val;
	int i;

	LOG(1, "INIT DAC\n");
	if (chip->type == SBXFI_HW_055x) {
		gpio = sbxfi_read(chip, GPIO);
		gpio &= ~GPIO_O_055X_LINEOUT1_MUTE;	/* unmute line-out1 */
		gpio |= GPIO_O_055X_ANALOG_IN_MUTE;	/* unmute analog-ins */
		sbxfi_write(chip, GPIO, gpio);
		return;
	}

	/* Mute outputs */
	gpio = sbxfi_read(chip, GPIO);
	gpio &= ~GPIO_O_BACK_LINEOUT_MUTE;
	sbxfi_write(chip, GPIO, gpio);

	sbxfi_reset_dac(chip);

	switch (chip->base_rate) {
	case 48000:
	default:
		val = 0x2400;
		break;
	case 96000:
		val = 0x2500;
		break;
	case 192000:
		val = 0x2600;
		break;
	}

	for (i = 0x06; i <= 0x0f; i += 3)
		sbxfi_set_dac_cmd(chip, val | i);

	/* unmute outputs */
	gpio = sbxfi_read(chip, GPIO);
	gpio |= GPIO_O_BACK_LINEOUT_MUTE;
	sbxfi_write(chip, GPIO, gpio);
}

static void sbxfi_setup_i2s(struct sbxfi *chip)
{
	unsigned int i2s;

	LOG(1, "SETUP I2S\n");
	i2s = sbxfi_read(chip, I2SCTL);
	i2s = i2s | 0x04040404; /* All I2S outputs enabled */
	/* 0x04 I2SA_L, 0x400 B_L, 0x40000 C_L, 0x4000000 D_L ?? */

	i2s &= 0xfffffffc;
	if (chip->base_rate == 96000)
		i2s |= 0x2;
	else /* 48000 */
		i2s |= 0x01;

	/* enable I2S-D input for capture */
	i2s |= 0x90000000;

	sbxfi_write(chip, I2SCTL, i2s);
}

static void sbxfi_set_adc_cmd(struct sbxfi *chip, unsigned int cmd)
{
	sbxfi_write_pci(chip, 0xe0, 0x001a0080);
	sbxfi_write_pci(chip, 0xe4, cmd);
	sbxfi_sync_i2c(chip);
}

static void setup_sb055x_adc(struct sbxfi *chip, int capsrc, int micboost)
{
	unsigned int val;

	val = sbxfi_read(chip, GPIO);
	val &= ~(GPIO_O_055X_AUX_MICLINE_SEL |
		 GPIO_O_055X_MIC_SW |
		 GPIO_O_055X_DIGITAL_IO_SEL |
		 GPIO_O_055X_MIC_POWER_ON |
		 GPIO_O_055X_ADC_BOOST);

	switch (capsrc) {
	case ADC_SRC_MICIN:
		val |= GPIO_O_055X_DIGITAL_IO_SEL;
		val |= GPIO_O_055X_MIC_SW;
		val |= GPIO_O_055X_MIC_POWER_ON;
		if (micboost)
			val |= GPIO_O_055X_ADC_BOOST;
		break;
	case ADC_SRC_LINEIN:
		val |= GPIO_O_055X_DIGITAL_IO_SEL;
		break;
	case ADC_SRC_AUX:
		val |= GPIO_O_055X_DIGITAL_IO_SEL;
		val |= GPIO_O_055X_AUX_MICLINE_SEL;
		break;
	case ADC_SRC_DIG:
	default:
		break;
	}
	sbxfi_write(chip, GPIO, val);
}

static void setup_adc(struct sbxfi *chip, int capsrc, int micboost)
{
	unsigned int input_source;
	unsigned int adcdata = 0;
	unsigned int gpio;

	input_source = GPIO_O_DIGITAL_IO_SELECT;	/* default to analog */
	switch (capsrc) {
	case ADC_SRC_MICIN:
		adcdata = 0x1;
		input_source |= GPIO_O_MIC_LINE_SELECT;
		break;
	case ADC_SRC_LINEIN:
		adcdata = 0x2;
		break;
	case ADC_SRC_VIDEO:
		adcdata = 0x4;
		break;
	case ADC_SRC_AUX:
		adcdata = 0x8;
		break;
	case ADC_SRC_DIG:
		adcdata = 0x0;
		input_source = 0;	/* set to Digital */
		break;
	}

	sbxfi_prepare_init(chip);

	sbxfi_set_adc_cmd(chip, 0x080e);
	sbxfi_set_adc_cmd(chip, 0x0a18);
	sbxfi_set_adc_cmd(chip, micboost ? 0xf71c : 0xcf1c);
	sbxfi_set_adc_cmd(chip, micboost ? 0xf71e : 0xcf1e);
	sbxfi_set_adc_cmd(chip, 0x8628);
	sbxfi_set_adc_cmd(chip, 0x2a | (adcdata << 0x8));

	gpio = sbxfi_read(chip, GPIO);
	gpio &= ~(GPIO_O_DIGITAL_IO_SELECT | GPIO_O_MIC_LINE_SELECT);
	gpio |= input_source;
	sbxfi_write(chip, GPIO, gpio);
	sbxfi_sync_i2c(chip);

	if (!(sbxfi_read(chip, CHIP_ID0) & 0x100))
		sbxfi_set_adc_cmd(chip, 0x2616);
}

static void sbxfi_init_adc(struct sbxfi *chip, int capsrc, int micboost)
{
	if (chip->type == SBXFI_HW_055x)
		setup_sb055x_adc(chip, capsrc, micboost);
	else
		setup_adc(chip, capsrc, micboost);
}

/*
 * TLB page management
 */

/* clear TLB page entries from the given idx */
static void sbxfi_free_tlb_pages(struct sbxfi *chip, unsigned int idx,
				 unsigned int pages)
{
	u32 *pgtbl = (u32*)chip->tlb.area + idx;

	/* Oh stupid bitmap ops */
	for (; pages--; idx++, pgtbl++) {
		clear_bit(idx, chip->tlb_page_bitmap);
		*pgtbl = chip->silent_addr;
	}
}

/* allocate TLB page entries for the given pages */
static int sbxfi_alloc_tlb_pages(struct sbxfi *chip, unsigned int pages)
{
	int idx, order = order_base_2(pages);

	idx = bitmap_find_free_region(chip->tlb_page_bitmap, SBXFI_TLB_PAGES,
				      order);
	if (idx < 0)
		return -1;
	/* release the rest tail */
	sbxfi_free_tlb_pages(chip, idx + pages, (1 << order) - pages);
	LOG(1, "allocated TLB at %d for %d pages\n", idx, pages);
	return idx;
}

/* clear the pages belonging to the port */
static void sbxfi_clear_tlb(struct sbxfi *chip, struct sbxfi_port *port)
{
	if (!port->tlb_pages)
		return;
	mutex_lock(&chip->tlb_mutex);
	LOG(1, "release TLB at %d for %d pages\n",
	    port->tlb_index, port->tlb_pages);
	sbxfi_free_tlb_pages(chip, port->tlb_index, port->tlb_pages);
	chip->used_pages -= port->tlb_pages;
	port->tlb_pages = 0;
	port->tlb_index = -1;
	if (!chip->used_pages) {
		LOG(1, "Disabling TLB buffer\n");
		sbxfi_write(chip, TRANS_TLB_PHY_ADDR_LO_0, 0);
		sbxfi_write(chip, TRANS_TLB_PHY_ADDR_HI_0, 0);
	}
	mutex_unlock(&chip->tlb_mutex);
}

/* allocate the pages belonging to the port */
static int sbxfi_setup_tlb(struct sbxfi *chip, struct sbxfi_port *port,
			   unsigned int bytes)
{
	struct snd_pcm_substream *substream = port->substream;
	unsigned int p, pages;
	int err = 0;
	u32 *pgtbl;

	pages = (bytes + SBXFI_PAGE_SIZE - 1) / SBXFI_PAGE_SIZE;

	mutex_lock(&chip->tlb_mutex);
	port->tlb_index = sbxfi_alloc_tlb_pages(chip, pages);
	if (port->tlb_index < 0) {
		printk(KERN_ERR PFX "Cannot allocate %d pages\n", pages);
		err = -ENOMEM;
		goto out;
	}
	port->tlb_pages = pages;

	pgtbl = (u32*)chip->tlb.area;
	pgtbl += port->tlb_index;
	for (p = 0; p < pages; p++, pgtbl++)
		*pgtbl = snd_pcm_sgbuf_get_addr(substream, p * SBXFI_PAGE_SIZE);

	if (!chip->used_pages) {
		LOG(1, "Setting TLB buffer page 0x%lx\n",
		    (unsigned long)chip->tlb.addr);
		sbxfi_write(chip, TRANS_TLB_PHY_ADDR_LO_0, (u32)chip->tlb.addr);
		sbxfi_write(chip, TRANS_TLB_PHY_ADDR_HI_0,
			    upper_32_bits(chip->tlb.addr));
	}
	chip->used_pages += pages;

 out:
	mutex_unlock(&chip->tlb_mutex);

	return err;
}

/*
 * port assignment
 */

static void sbxfi_port_volume_init(struct sbxfi_port *port);

static struct sbxfi_port *sbxfi_port_alloc(struct sbxfi *chip,
					   struct snd_pcm_substream *substream)
{
	struct sbxfi_port *port;
	int src;

	port = kzalloc(sizeof(*port), GFP_KERNEL);
	if (!port)
		return NULL;
	port->tlb_index = -1;
	port->substream = substream;
	port->state = PORT_OPENED;
	sbxfi_port_volume_init(port);

	/* find a pair hole (order = 1) */
	src = bitmap_find_free_region(chip->src_bitmap, SBXFI_MAX_SRCS, 1);
	if (src < 0) {
		kfree(port);
		return NULL;
	}
	port->src[0] = src;
	port->src[1] = src + 1;
	LOG(1, "Allocate SRC %d\n", src);
	spin_lock_irq(&chip->port_lock);
	if (list_empty(&chip->port_list)) {
		/* Enable SRC input from Audio Ring */
		sbxfi_write(chip, SRC_MASTER_CTL, 0x1);
	}
	list_add(&port->list, &chip->port_list);
	spin_unlock_irq(&chip->port_lock);
	return port;
}

static void sbxfi_port_free(struct sbxfi *chip, struct sbxfi_port *port)
{
	if (!port)
		return;
	spin_lock_irq(&chip->port_lock);
	list_del(&port->list);
	if (list_empty(&chip->port_list)) {
		/* Disable SRC inputs from Audio Ring */
		sbxfi_write(chip, SRC_MASTER_CTL, 0x0);
	}
	spin_unlock_irq(&chip->port_lock);
	LOG(1, "Release SRC %d\n", port->src[0]);
	bitmap_release_region(chip->src_bitmap, port->src[0], 1);
	kfree(port);
}

/*
 */

/* clear input mappers */
static void sbxfi_daoimap_clear(struct sbxfi *chip)
{
	int i;
	LOG(2, "DAOIMAP CLEAR\n");
	for (i = 0; i < 0x50; i++)
		sbxfi_write(chip, DAOIMAP_START(i), 0);
}

/* Input mapper is a circular linked-list */
static void sbxfi_setup_play_mapper(struct sbxfi *chip,
				    struct sbxfi_port *port)
{
	struct snd_pcm_runtime *runtime = port->substream->runtime;
	unsigned int dio1 = port->conv[0];
	unsigned int dio2 = port->conv[1];

	sbxfi_daoimap_clear(chip);
	if (runtime->rate <= 48000) {
		sbxfi_write(chip, DAOIMAP_START(dio1), 0);
		sbxfi_write(chip, DAOIMAP_START(dio2),
			    (dio1 << 16) |
			    sbxfi_get_audio_src_chan(port->src[1]));
		sbxfi_write(chip, DAOIMAP_START(dio1),
			    (dio2 << 16) |
			    sbxfi_get_audio_src_chan(port->src[0]));
	} else {
		sbxfi_write(chip, DAOIMAP_START(dio1), 0);
		sbxfi_write(chip, DAOIMAP_START(dio2),
			    ((dio1 + 2) << 16) |
			    sbxfi_get_audio_src_chan(port->src[1]));
		sbxfi_write(chip, DAOIMAP_START(dio1 + 2),
			    ((dio2 + 2) << 16) |
			    sbxfi_get_audio_src_chan(port->src[0] + 0x80));
		sbxfi_write(chip, DAOIMAP_START(dio2 + 2),
			    (dio1 << 16) |
			    sbxfi_get_audio_src_chan(port->src[1] + 0x80));
		sbxfi_write(chip, DAOIMAP_START(dio1),
			    (dio2 << 16) |
			    sbxfi_get_audio_src_chan(port->src[0]));
	}
}

static void sbxfi_cleanup_play_mapper(struct sbxfi *chip,
				      struct sbxfi_port *port)
{
	sbxfi_daoimap_clear(chip);
}

static void sbxfi_setup_play_pitch(struct sbxfi *chip,
				   struct sbxfi_port *port)
{
	struct snd_pcm_runtime *runtime = port->substream->runtime;
	int i;
	unsigned int pitch, pitch_ch;

	/* convert to fixed-point 8.24 format */
	/* pitch = div_u64(runtime->rate * (1ULL << 24), chip->base_rate); */
	pitch = (1 << 24); /* currently fixed rate */


	for (i = 0; i < 2; i++) {
		/* write the pitch to param ring of the corresponsing SRC
		 * pitch slot
		 */
		pitch_ch = sbxfi_get_param_pitch_chan(port->src[i]);
		LOG(1, "Pitch [%0x:%0x] = 0x%x\n", port->src[i], pitch_ch,
		    pitch);
		sbxfi_write(chip, PRING_LO_HI_START(pitch_ch), pitch);

		if (runtime->rate <= 48000)
			continue;

		/* set up conjugate channel */
		pitch_ch = sbxfi_get_param_pitch_chan(port->src[i] + 0x80);
		LOG(1, "Pitch [%0x:%0x] = 0x%x\n", port->src[i] + 0x80,
		    pitch_ch, pitch);
		sbxfi_write(chip, PRING_LO_HI_START(pitch_ch), pitch);
	}
}

#define MAX_VOLUME	144	/* Centibel steps */
#define AMOP_0dB	0x1c00	/* 0dB */

static const DECLARE_TLV_DB_SCALE(db_amop_vol, -2880, 20, 1);

/* linear -> dB conversion table */
static unsigned short amop_vol_table[MAX_VOLUME + 1] = {
	0x0000, 0x010a, 0x0110, 0x0116, 0x011d,
	0x0124, 0x012a, 0x0131, 0x0138, 0x0140,
	0x0147, 0x014f, 0x0157, 0x015f, 0x0167,
	0x016f, 0x0178, 0x0180, 0x0189, 0x0193,
	0x019c, 0x01a6, 0x01af, 0x01b9, 0x01c4,
	0x01ce, 0x01d9, 0x01e4, 0x01ef, 0x01fb,
	0x0207, 0x0213, 0x021f, 0x022c, 0x0239,
	0x0246, 0x0254, 0x0262, 0x0270, 0x027e,
	0x028d, 0x029c, 0x02ac, 0x02bc, 0x02cc,
	0x02dd, 0x02ee, 0x0300, 0x0311, 0x0324,
	0x0336, 0x034a, 0x035d, 0x0371, 0x0386,
	0x039b, 0x03b0, 0x03c6, 0x03dd, 0x03f4,
	0x040c, 0x0424, 0x043c, 0x0456, 0x0470,
	0x048a, 0x04a5, 0x04c1, 0x04dd, 0x04fa,
	0x0518, 0x0536, 0x0555, 0x0575, 0x0596,
	0x05b7, 0x05d9, 0x05fc, 0x0620, 0x0644,
	0x066a, 0x0690, 0x06b7, 0x06df, 0x0708,
	0x0732, 0x075d, 0x0789, 0x07b6, 0x07e4,
	0x0813, 0x0843, 0x0874, 0x08a7, 0x08da,
	0x090f, 0x0945, 0x097c, 0x09b5, 0x09ef,
	0x0a2a, 0x0a67, 0x0aa5, 0x0ae4, 0x0b25,
	0x0b68, 0x0bac, 0x0bf1, 0x0c38, 0x0c81,
	0x0ccc, 0x0d18, 0x0d66, 0x0db6, 0x0e08,
	0x0e5c, 0x0eb1, 0x0f09, 0x0f63, 0x0fbe,
	0x101c, 0x107c, 0x10df, 0x1143, 0x11aa,
	0x1214, 0x127f, 0x12ee, 0x135f, 0x13d2,
	0x1448, 0x14c1, 0x153d, 0x15bc, 0x163d,
	0x16c2, 0x174a, 0x17d4, 0x1863, 0x18f4,
	0x1989, 0x1a21, 0x1abd, 0x1b5c, 0x1c00
};

static void sbxfi_port_volume_init(struct sbxfi_port *port)
{
	port->vol[0] = port->vol[1] = MAX_VOLUME;
}

static void sbxfi_setup_play_mixer(struct sbxfi *chip,
				   struct sbxfi_port *port)
{
	struct snd_pcm_runtime *runtime = port->substream->runtime;
	int i, vol, audio_ch;

	for (i = 0; i < 2; i++) {
		vol = port->vol[i];
		vol += chip->vol[VOLUME_MASTER][i] - MAX_VOLUME;
		vol += chip->vol[VOLUME_PCM][i] - MAX_VOLUME;
		if (vol < 0)
			vol = 0;
		vol = amop_vol_table[vol];
		audio_ch = sbxfi_get_audio_src_chan(port->src[i]);
		LOG(1, "Amp [%02x:%04x] = 0x%x\n", port->src[i],
		    audio_ch, vol);
		sbxfi_write_amop(chip, audio_ch, vol, audio_ch, 0);
		if (runtime->rate <= 48000)
			continue;

		audio_ch = sbxfi_get_audio_src_chan(port->src[i] + 0x80);
		LOG(1, "Amp [%02x:%04x] = 0x%x\n", port->src[i] + 0x80,
		    audio_ch, vol);
		sbxfi_write_amop(chip, audio_ch, vol, audio_ch, 0);
	}
}

static void sbxfi_setup_capture_mapper(struct sbxfi *chip,
				       struct sbxfi_port *port)
{
	struct snd_pcm_runtime *runtime = port->substream->runtime;
	unsigned int srcch = port->src[0];
	unsigned int srcch2 = port->src[1];

	if (runtime->rate <= 48000) {
		sbxfi_write(chip, SRCIMAP(srcch2), 0);
		sbxfi_write(chip, SRCIMAP(srcch),
			    (srcch << 24) |
			    (srcch2 << 16) |
			    sbxfi_get_audio_sum_chan(srcch));
		sbxfi_write(chip, SRCIMAP(srcch2),
			    (srcch2 << 24) |
			    (srcch << 16) |
			    sbxfi_get_audio_sum_chan(srcch2));
	} else {
		sbxfi_write(chip, SRCIMAP(srcch2 + 0x80), 0);
		sbxfi_write(chip, SRCIMAP(srcch),
			    (srcch << 24) |
			    (srcch2 << 16) |
			    sbxfi_get_audio_sum_chan(srcch));
		sbxfi_write(chip, SRCIMAP(srcch2),
			    (srcch2 << 24) |
			    ((srcch + 0x80) << 16) |
			    sbxfi_get_audio_sum_chan(srcch2));
		sbxfi_write(chip, SRCIMAP(srcch + 0x80),
			    ((srcch + 0x80) << 24) |
			    ((srcch2 + 0x80) << 16) |
			    sbxfi_get_audio_sum_chan(srcch + 0x80));
		sbxfi_write(chip, SRCIMAP(srcch2 + 0x80),
			    ((srcch2 + 0x80) << 24) |
			    (srcch << 16) |
			    sbxfi_get_audio_sum_chan(srcch2 + 0x80));
	}
}

#define sbxfi_setup_capture_pitch	sbxfi_setup_play_pitch

static void sbxfi_setup_capture_mixer(struct sbxfi *chip,
				      struct sbxfi_port *port)
{
	unsigned int srcch = port->src[0];
	unsigned int srcch2 = port->src[1];
	unsigned int *vol = &chip->vol[VOLUME_CAPTURE][0];
	unsigned int inch1, inch2;
	int i;

	/* I2S-In3 */
	inch1 = DAI_CH_I2SIND1L;
	inch2 = DAI_CH_I2SIND1R;
	/* program all I2S-In3 slots */
	for (i = 0; i < 8; i++) {
		if (i <= 3) {
			sbxfi_write_amop(chip, inch1 + (i * 0x200), vol[0],
					 inch1 + (i * 0x200),
					 (0x80000000 + srcch));
			sbxfi_write_amop(chip, inch2 + (i * 0x200), vol[1],
					 inch2 + (i * 0x200),
					 (0x80000000 + srcch2));
		} else {
			sbxfi_write_amop(chip, inch1 + (i * 0x200), vol[0],
					 inch1 + (i * 0x200),
					 (0x80000000 + srcch + 0x80));
			sbxfi_write_amop(chip, inch2 + (i * 0x200), vol[1],
					 inch2 + (i * 0x200),
					 (0x80000000 + srcch2 + 0x80));
		}
	}
}

static int sbxfi_update_mixer(struct sbxfi *chip, int *vol, int type)
{
	int i, changed = 0;
	int *refvol = &chip->vol[type][0];
	struct sbxfi_port *port;

	if (vol[0] < 0)
		vol[0] = 0;
	else if (vol[0] > MAX_VOLUME)
		vol[0] = MAX_VOLUME;
	if (vol[1] < 0)
		vol[1] = 0;
	else if (vol[1] > MAX_VOLUME)
		vol[1] = MAX_VOLUME;

	for (i = 0; i < 2; i++) {
		if (vol[i] == refvol[i])
			continue;
		refvol[i] = vol[i];
		spin_lock_irq(&chip->port_lock);
		list_for_each_entry(port, &chip->port_list, list) {
			if (port->state < PORT_PREPARED)
				continue;
			if (type == VOLUME_CAPTURE) {
				if (port->substream->stream !=
				    SNDRV_PCM_STREAM_CAPTURE)
					continue;
				sbxfi_setup_capture_mixer(chip, port);
			} else {
				if (port->substream->stream !=
				    SNDRV_PCM_STREAM_PLAYBACK)
					continue;
				sbxfi_setup_play_mixer(chip, port);
			}
		}
		spin_unlock_irq(&chip->port_lock);
		changed = 1;
	}
	return changed;
}


/*
 * PCM callbacks
 */
static struct snd_pcm_hardware sbxfi_pcm_hardware = {
	.info =		(SNDRV_PCM_INFO_MMAP |
			 SNDRV_PCM_INFO_INTERLEAVED |
			 SNDRV_PCM_INFO_BLOCK_TRANSFER |
			 SNDRV_PCM_INFO_MMAP_VALID),
	.formats =	SNDRV_PCM_FMTBIT_S16_LE,
	.rates =	SNDRV_PCM_RATE_CONTINUOUS | SNDRV_PCM_RATE_8000_96000,
	.rate_min =	8000,
	.rate_max =	96000,
	.channels_min =	2,
	.channels_max =	2,
	.buffer_bytes_max =	SBXFI_MAX_BUFFER,
	.period_bytes_min =	64,
	.period_bytes_max =	SBXFI_MAX_BUFFER,
	.periods_min =		1,
	.periods_max =		4096,
};

static void set_rate_limit(struct sbxfi *chip, struct snd_pcm_runtime *runtime)
{
#ifndef XXX_CONT_RATE
	switch (chip->base_rate) {
	case 96000:
		runtime->hw.rates = SNDRV_PCM_RATE_96000;
		break;
	default:
		runtime->hw.rates = SNDRV_PCM_RATE_48000;
		break;
	}
	runtime->hw.rate_min = runtime->hw.rate_max = chip->base_rate;
#endif
}

static int sbxfi_playback_open(struct snd_pcm_substream *substream)
{
	struct sbxfi *chip = snd_pcm_substream_chip(substream);
	struct snd_pcm_runtime *runtime = substream->runtime;
	struct sbxfi_port *port;

	port = sbxfi_port_alloc(chip, substream);
	if (!port)
		return -EBUSY;
	runtime->private_data = port;
	port->conv[0] = I2SA_L;
	port->conv[1] = I2SA_R;
#if 0 /* XXX */
	port->conv[2] = I2SB_L;
	port->conv[3] = I2SB_R;
	port->conv[4] = I2SC_L;
	port->conv[5] = I2SC_R;
#endif
	runtime->hw = sbxfi_pcm_hardware;
	set_rate_limit(chip, runtime);
	snd_pcm_hw_constraint_integer(runtime, SNDRV_PCM_HW_PARAM_PERIODS);
	snd_pcm_hw_constraint_step(runtime, 0, SNDRV_PCM_HW_PARAM_BUFFER_BYTES,
				   128);
	snd_pcm_hw_constraint_step(runtime, 0, SNDRV_PCM_HW_PARAM_PERIOD_BYTES,
				   128);
	return 0;
}

static int sbxfi_playback_close(struct snd_pcm_substream *substream)
{
	struct sbxfi *chip = snd_pcm_substream_chip(substream);
	struct sbxfi_port *port = substream->runtime->private_data;
	sbxfi_cleanup_play_mapper(chip, port);
	sbxfi_port_free(chip, port);
	return 0;
}

static int sbxfi_capture_open(struct snd_pcm_substream *substream)
{
	struct sbxfi *chip = snd_pcm_substream_chip(substream);
	struct snd_pcm_runtime *runtime = substream->runtime;
	struct sbxfi_port *port;

	port = sbxfi_port_alloc(chip, substream);
	if (!port)
		return -EBUSY;
	runtime->private_data = port;
	runtime->hw = sbxfi_pcm_hardware;
	set_rate_limit(chip, runtime);
	snd_pcm_hw_constraint_integer(runtime, SNDRV_PCM_HW_PARAM_PERIODS);
	snd_pcm_hw_constraint_step(runtime, 0, SNDRV_PCM_HW_PARAM_BUFFER_BYTES,
				   128);
	snd_pcm_hw_constraint_step(runtime, 0, SNDRV_PCM_HW_PARAM_PERIOD_BYTES,
				   128);
	return 0;
}

static int sbxfi_capture_close(struct snd_pcm_substream *substream)
{
	struct sbxfi *chip = snd_pcm_substream_chip(substream);
	struct sbxfi_port *port = substream->runtime->private_data;
	sbxfi_port_free(chip, port);
	return 0;
}

static int sbxfi_pcm_hw_params(struct snd_pcm_substream *substream,
			       struct snd_pcm_hw_params *hw_params)
{
	struct sbxfi *chip = snd_pcm_substream_chip(substream);
	struct sbxfi_port *port = substream->runtime->private_data;
	unsigned int bytes = params_buffer_bytes(hw_params);
	int err;

	sbxfi_clear_tlb(chip, port);
	err = snd_pcm_lib_malloc_pages(substream, bytes);
	if (err < 0)
		return err;
	return sbxfi_setup_tlb(chip, port, bytes);
}

static int sbxfi_pcm_hw_free(struct snd_pcm_substream *substream)
{
	struct sbxfi *chip = snd_pcm_substream_chip(substream);
	struct sbxfi_port *port = substream->runtime->private_data;

	sbxfi_clear_tlb(chip, port);
	return snd_pcm_lib_free_pages(substream);
}

static int sbxfi_playback_prepare(struct snd_pcm_substream *substream)
{
	struct snd_pcm_runtime *runtime = substream->runtime;
	struct sbxfi *chip = snd_pcm_substream_chip(substream);
	struct sbxfi_port *port = runtime->private_data;

	LOG(1, "PLAYBACK PREPARE: rate=%d, period_size=%d, buffer_size=%d\n",
	    runtime->rate, (int)runtime->period_size, (int)runtime->buffer_size);
	port->last_pos = 0;
	port->frag_count = runtime->period_size;
	port->need_update = 0;

	port->start_addr = port->tlb_index * SBXFI_PAGE_SIZE;
	port->buffer_bytes = frames_to_bytes(runtime, runtime->buffer_size);

	sbxfi_setup_play_pitch(chip, port);
	sbxfi_setup_play_mixer(chip, port);
	sbxfi_setup_play_mapper(chip, port);

	port->state = PORT_PREPARED;
	return 0;
}

static int sbxfi_capture_prepare(struct snd_pcm_substream *substream)
{
	struct snd_pcm_runtime *runtime = substream->runtime;
	struct sbxfi *chip = snd_pcm_substream_chip(substream);
	struct sbxfi_port *port = runtime->private_data;

	port->last_pos = 0;
	port->frag_count = runtime->period_size;
	port->need_update = 0;

	port->start_addr = port->tlb_index * SBXFI_PAGE_SIZE;
	port->buffer_bytes = frames_to_bytes(runtime, runtime->buffer_size);

	sbxfi_init_adc(chip, chip->capsrc, chip->micboost);

	sbxfi_setup_capture_mapper(chip, port);
	sbxfi_setup_capture_pitch(chip, port);
	sbxfi_setup_capture_mixer(chip, port);

	port->state = PORT_PREPARED;
	return 0;
}

static int sbxfi_playback_trigger(struct snd_pcm_substream *substream, int cmd)
{
	struct sbxfi *chip = snd_pcm_substream_chip(substream);
	struct sbxfi_port *port = substream->runtime->private_data;

	switch (cmd) {
	case SNDRV_PCM_TRIGGER_START:
		LOG(1, "PLAY TRIGGER START\n");
		sbxfi_playback_start(chip, port);
		port->state = PORT_RUNNING;
		break;
	case SNDRV_PCM_TRIGGER_STOP:
		LOG(1, "PLAY TRIGGER STOP\n");
		port->state = PORT_PREPARED;
		sbxfi_playback_stop(chip, port);
		break;
	}
	LOG(1, "PLAY UPDATE TIMER\n");
	sbxfi_update_timer(chip);
	return 0;
}

static int sbxfi_capture_trigger(struct snd_pcm_substream *substream, int cmd)
{
	struct sbxfi *chip = snd_pcm_substream_chip(substream);
	struct sbxfi_port *port = substream->runtime->private_data;

	switch (cmd) {
	case SNDRV_PCM_TRIGGER_START:
		sbxfi_capture_start(chip, port);
		port->state = PORT_RUNNING;
		break;
	case SNDRV_PCM_TRIGGER_STOP:
		port->state = PORT_PREPARED;
		sbxfi_capture_stop(chip, port);
		break;
	}
	sbxfi_update_timer(chip);
	return 0;
}

static snd_pcm_uframes_t
sbxfi_pcm_pointer(struct snd_pcm_substream *substream)
{
	struct sbxfi *chip = snd_pcm_substream_chip(substream);
	struct snd_pcm_runtime *runtime = substream->runtime;
	struct sbxfi_port *port = runtime->private_data;
	unsigned int pos;

	pos = sbxfi_read(chip, SRCCA(port->src[0])) & 0x03ffffff;
	pos -= port->start_addr;
	/* The pointer is always 128 bytes ahead */
	if (pos >= CA_OFFSET)
		pos -= CA_OFFSET;
	else
		pos = pos + port->buffer_bytes - CA_OFFSET;
	pos = bytes_to_frames(runtime, pos);
	pos %= runtime->buffer_size;
	LOG(2, "POINTER = 0x%x\n", pos);
	return pos;
}

/*
 * PCM registration
 */
static struct snd_pcm_ops sbxfi_playback_ops = {
	.open = sbxfi_playback_open,
	.close = sbxfi_playback_close,
	.ioctl = snd_pcm_lib_ioctl,
	.hw_params = sbxfi_pcm_hw_params,
	.hw_free = sbxfi_pcm_hw_free,
	.prepare = sbxfi_playback_prepare,
	.trigger = sbxfi_playback_trigger,
	.pointer = sbxfi_pcm_pointer,
	.page = snd_pcm_sgbuf_ops_page,
};

static struct snd_pcm_ops sbxfi_capture_ops = {
	.open = sbxfi_capture_open,
	.close = sbxfi_capture_close,
	.ioctl = snd_pcm_lib_ioctl,
	.hw_params = sbxfi_pcm_hw_params,
	.hw_free = sbxfi_pcm_hw_free,
	.prepare = sbxfi_capture_prepare,
	.trigger = sbxfi_capture_trigger,
	.pointer = sbxfi_pcm_pointer,
	.page = snd_pcm_sgbuf_ops_page,
};

static int __devinit sbxfi_create_pcm(struct sbxfi *chip)
{
	struct snd_pcm *pcm;
	int err;

	err = snd_pcm_new(chip->card, "SB-XFi", 0, 1, 1, &pcm);
	if (err < 0)
		return err;
	snd_pcm_set_ops(pcm, SNDRV_PCM_STREAM_PLAYBACK, &sbxfi_playback_ops);
	snd_pcm_set_ops(pcm, SNDRV_PCM_STREAM_CAPTURE, &sbxfi_capture_ops);
	pcm->private_data = chip;
	pcm->info_flags = 0;
	strcpy(pcm->name, "SB-XFi");
	snd_pcm_lib_preallocate_pages_for_all(pcm, SNDRV_DMA_TYPE_DEV_SG,
					      snd_dma_pci_data(chip->pci),
					      1024 * 64, 32 * 1024 * 1024);
	chip->pcm = pcm;
	return 0;
}

/*
 * Mixer interface
 */
static int sbxfi_chip_vol_info(struct snd_kcontrol *kcontrol,
			       struct snd_ctl_elem_info *uinfo)
{
	uinfo->type = SNDRV_CTL_ELEM_TYPE_INTEGER;
	uinfo->count = 2;
	uinfo->value.integer.min = 0;
	uinfo->value.integer.max = MAX_VOLUME;
	return 0;
}

static int sbxfi_chip_vol_get(struct snd_kcontrol *kcontrol,
			      struct snd_ctl_elem_value *ucontrol)
{
	struct sbxfi *chip = snd_kcontrol_chip(kcontrol);
	int type = kcontrol->private_value;
	ucontrol->value.integer.value[0] = chip->vol[type][0];
	ucontrol->value.integer.value[1] = chip->vol[type][1];
	return 0;
}

static int sbxfi_chip_vol_put(struct snd_kcontrol *kcontrol,
			      struct snd_ctl_elem_value *ucontrol)
{
	struct sbxfi *chip = snd_kcontrol_chip(kcontrol);
	int type = kcontrol->private_value;
	int vol[2];

	vol[0] = ucontrol->value.integer.value[0];
	vol[1] = ucontrol->value.integer.value[1];
	return sbxfi_update_mixer(chip, vol, type);
}

/*
 */

static int sb055x_adcs[] = {
	ADC_SRC_MICIN, ADC_SRC_LINEIN, ADC_SRC_AUX, ADC_SRC_DIG
};
static int sbxfi_adcs[] = {
	ADC_SRC_MICIN, ADC_SRC_LINEIN, ADC_SRC_AUX, ADC_SRC_VIDEO, ADC_SRC_DIG
};

static int sbxfi_capsrc_info(struct snd_kcontrol *kcontrol,
			     struct snd_ctl_elem_info *uinfo)
{
	struct sbxfi *chip = snd_kcontrol_chip(kcontrol);
	static char *texts[ADC_SRC_NUMS] = {
		"Mic", "Line", "Aux", "Video", "Digital",
	};
	int *adcs, nums;

	uinfo->type = SNDRV_CTL_ELEM_TYPE_ENUMERATED;
	uinfo->count = 1;
	if (chip->type == SBXFI_HW_055x) {
		adcs = sb055x_adcs;
		nums = ARRAY_SIZE(sb055x_adcs);
	} else {
		adcs = sbxfi_adcs;
		nums = ARRAY_SIZE(sbxfi_adcs);
	}
	uinfo->value.enumerated.items = nums;
	if (uinfo->value.enumerated.item >= nums)
		uinfo->value.enumerated.item = nums - 1;
	strcpy(uinfo->value.enumerated.name,
	       texts[adcs[uinfo->value.enumerated.item]]);
	return 0;
}

static int sbxfi_capsrc_get(struct snd_kcontrol *kcontrol,
			    struct snd_ctl_elem_value *ucontrol)
{
	struct sbxfi *chip = snd_kcontrol_chip(kcontrol);
	int i, *adcs, nums;
	if (chip->type == SBXFI_HW_055x) {
		adcs = sb055x_adcs;
		nums = ARRAY_SIZE(sb055x_adcs);
	} else {
		adcs = sbxfi_adcs;
		nums = ARRAY_SIZE(sbxfi_adcs);
	}
	for (i = 0; i < nums; i++) {
		if (adcs[i] == chip->capsrc) {
			ucontrol->value.enumerated.item[0] = i;
			break;
		}
	}
	return 0;
}

static int sbxfi_capsrc_put(struct snd_kcontrol *kcontrol,
			    struct snd_ctl_elem_value *ucontrol)
{
	struct sbxfi *chip = snd_kcontrol_chip(kcontrol);
	int capsrc, *adcs, nums;
	if (chip->type == SBXFI_HW_055x) {
		adcs = sb055x_adcs;
		nums = ARRAY_SIZE(sb055x_adcs);
	} else {
		adcs = sbxfi_adcs;
		nums = ARRAY_SIZE(sbxfi_adcs);
	}
	capsrc = ucontrol->value.enumerated.item[0];
	if (capsrc >= nums)
		capsrc = nums - 1;
	capsrc = adcs[capsrc];
	if (capsrc == chip->capsrc)
		return 0;
	chip->capsrc = capsrc;
	sbxfi_init_adc(chip, chip->capsrc, chip->micboost);
	return 1;
}

#define sbxfi_micboost_info	snd_ctl_boolean_mono_info

static int sbxfi_micboost_get(struct snd_kcontrol *kcontrol,
			      struct snd_ctl_elem_value *ucontrol)
{
	struct sbxfi *chip = snd_kcontrol_chip(kcontrol);
	ucontrol->value.integer.value[0] = chip->micboost;
	return 0;
}

static int sbxfi_micboost_put(struct snd_kcontrol *kcontrol,
			      struct snd_ctl_elem_value *ucontrol)
{
	struct sbxfi *chip = snd_kcontrol_chip(kcontrol);
	int boost = !!ucontrol->value.integer.value[0];
	if (chip->micboost == boost)
		return 0;
	chip->micboost = boost;
	sbxfi_init_adc(chip, chip->capsrc, chip->micboost);
	return 1;
}

static struct snd_kcontrol_new mixers[] __devinitdata = {
	{
		.name = "Master Playback Volume",
		.iface = SNDRV_CTL_ELEM_IFACE_MIXER,
		.access = (SNDRV_CTL_ELEM_ACCESS_READWRITE
			   | SNDRV_CTL_ELEM_ACCESS_TLV_READ),
		.info = sbxfi_chip_vol_info,
		.get = sbxfi_chip_vol_get,
		.put = sbxfi_chip_vol_put,
		.private_value = VOLUME_MASTER,
		.tlv = { .p = db_amop_vol }
	},
	{
		.name = "PCM Playback Volume",
		.iface = SNDRV_CTL_ELEM_IFACE_MIXER,
		.access = (SNDRV_CTL_ELEM_ACCESS_READWRITE
			   | SNDRV_CTL_ELEM_ACCESS_TLV_READ),
		.info = sbxfi_chip_vol_info,
		.get = sbxfi_chip_vol_get,
		.put = sbxfi_chip_vol_put,
		.private_value = VOLUME_PCM,
		.tlv = { .p = db_amop_vol }
	},
	{
		.name = "Capture Volume",
		.iface = SNDRV_CTL_ELEM_IFACE_MIXER,
		.access = (SNDRV_CTL_ELEM_ACCESS_READWRITE
			   | SNDRV_CTL_ELEM_ACCESS_TLV_READ),
		.info = sbxfi_chip_vol_info,
		.get = sbxfi_chip_vol_get,
		.put = sbxfi_chip_vol_put,
		.private_value = VOLUME_CAPTURE,
		.tlv = { .p = db_amop_vol }
	},
	{
		.name = "Capture Source",
		.iface = SNDRV_CTL_ELEM_IFACE_MIXER,
		.access = SNDRV_CTL_ELEM_ACCESS_READWRITE,
		.info = sbxfi_capsrc_info,
		.get = sbxfi_capsrc_get,
		.put = sbxfi_capsrc_put,
	},
	{
		.name = "Mic Boost",
		.iface = SNDRV_CTL_ELEM_IFACE_MIXER,
		.access = SNDRV_CTL_ELEM_ACCESS_READWRITE,
		.info = sbxfi_micboost_info,
		.get = sbxfi_micboost_get,
		.put = sbxfi_micboost_put,
	},
};

static int __devinit sbxfi_create_mixer(struct sbxfi *chip)
{
	int i, err;

	/* initial volume 0dB for PCM */
	chip->vol[VOLUME_PCM][0] = chip->vol[VOLUME_PCM][1] = MAX_VOLUME;
	/* master is muted */

	for (i = 0; i < ARRAY_SIZE(mixers); i++) {
		err = snd_ctl_add(chip->card, snd_ctl_new1(&mixers[i], chip));
		if (err < 0)
			return err;
	}
	return 0;
}

/*
 * proc interface
 */

static char *typenames[SBXFI_HW_TYPES] = {
	[SBXFI_HW_ORIG] = "SB0760",
	[SBXFI_HW_055x] = "SB0550",
	[SBXFI_HW_073x] = "SB0730",
	[SBXFI_HW_UAA] = "SB0886"
};

#ifdef CONFIG_PROC_FS
static void sbxfi_proc_read(struct snd_info_entry * entry,
			    struct snd_info_buffer *buffer)
{
	struct sbxfi *chip = entry->private_data;
	int i;

	snd_iprintf(buffer, "SBXFI SSID=%04x, type=%s\n",
		    chip->pci->subsystem_device, typenames[chip->type]);
	snd_iprintf(buffer, "GLOBAL_CTL = %x\n",
		    sbxfi_read(chip, GLOBAL_CTL));
	snd_iprintf(buffer, "PLL_CTL = %x\n",
		    sbxfi_read(chip, PLL_CTL));
	snd_iprintf(buffer, "GPIO_CTL = %x\n",
		    sbxfi_read(chip, GPIO_CTL));
	snd_iprintf(buffer, "GPIO = %x\n",
		    sbxfi_read(chip, GPIO));
	snd_iprintf(buffer, "I2SCTL = %x\n",
		    sbxfi_read(chip, I2SCTL));
	snd_iprintf(buffer, "GLOBAL_INT_ENABLE = %x\n",
		    sbxfi_read(chip, GLOBAL_INT_ENABLE));
	snd_iprintf(buffer, "\n");
	snd_iprintf(buffer, "DAOIMAP\n");
	for (i = 0; i < 0x50; i++) {
		snd_iprintf(buffer, "%02x:%08x ", i,
			    sbxfi_read(chip, DAOIMAP_START(i)));
		if (((i + 1) % 8) == 0)
			snd_iprintf(buffer, "\n");
	}
}

static void __devinit sbxfi_proc_init(struct sbxfi *chip)
{
	struct snd_info_entry *entry;

	if (!snd_card_proc_new(chip->card, "xfi", &entry))
		snd_info_set_text_ops(entry, chip, sbxfi_proc_read);
}
#else
#define sbxfi_proc_init(x)
#endif

/*
 * Initialization
 */

/* allocate TLB buffer */
static int sbxfi_alloc_buffer(struct sbxfi *chip)
{
	int p, err;
	void *sptr;
	u32 *pgtbl;

	err = snd_dma_alloc_pages(SNDRV_DMA_TYPE_DEV,
				  snd_dma_pci_data(chip->pci),
				  PAGE_SIZE, &chip->tlb);
	if (err < 0) {
		printk(KERN_ERR PFX "Cannot allocate TLB\n");
		return -ENOMEM;
	}

	sptr = (void *)get_zeroed_page(GFP_DMA32 | GFP_KERNEL);
	if (!sptr) {
		printk(KERN_ERR PFX "Cannot allocate a silent page\n");
		snd_dma_free_pages(&chip->tlb);
		return -ENOMEM;
	}

	/* fill silent pages */
	chip->silent_page = virt_to_page(sptr);
	chip->silent_addr = page_to_phys(chip->silent_page);
	pgtbl = (u32*)chip->tlb.area;
	for (p = 0; p < SBXFI_TLB_PAGES; p++, pgtbl++)
		*pgtbl = chip->silent_addr;

	return 0;
}

/* release TLB buffer */
static void sbxfi_free_buffer(struct sbxfi *chip)
{
	if (chip->tlb.area) {
		sbxfi_write(chip, TRANS_TLB_PHY_ADDR_LO_0, 0);
		sbxfi_write(chip, TRANS_TLB_PHY_ADDR_HI_0, 0);
		snd_dma_free_pages(&chip->tlb);
		__free_page(chip->silent_page);
	}
}

static int sbxfi_free(struct sbxfi *chip)
{
	if (chip->port)
		sbxfi_write(chip, GLOBAL_INT_ENABLE, 0); /* disable irqs */
	if (chip->irq >= 0) {
		synchronize_irq(chip->irq);
		free_irq(chip->irq, chip);
	}
	sbxfi_free_buffer(chip);
	pci_release_region(chip->pci, chip->bar);
	pci_disable_device(chip->pci);
	kfree(chip);
	return 0;
}


static int sbxfi_dev_free(struct snd_device *device)
{
	return sbxfi_free(device->device_data);
}

#define RESTORE_PCI(reg) pci_write_config_dword(chip->pci, reg, config[reg/4])

/* Switch to X-Fi mode */
static void sbxfi_switch_xfi_mode(struct sbxfi *chip)
{
	unsigned int config[16];
	int i;

	for (i = 0; i < 16; i++) {
		pci_read_config_dword(chip->pci, i * 4, &config[i]);
		LOG(1, "Before X-Fi mode: PCI[0x%x] = 0x%x\n", i, config[i]);
	}

	/* write magic word to switch to X-Fi mode */
	snd_printdd("Switching to X-Fi mode\n");
	pci_write_config_dword(chip->pci, 0xa0, 0x87654321);
	/* restore bar5 */
	RESTORE_PCI(PCI_BASE_ADDRESS_5);
	/* Switch to XFi core config space without BAR0 */
	pci_write_config_dword(chip->pci, 0xa0, 0x12345678);
	/* restore other BARs */
	RESTORE_PCI(PCI_BASE_ADDRESS_1);
	RESTORE_PCI(PCI_BASE_ADDRESS_2);
	RESTORE_PCI(PCI_BASE_ADDRESS_3);
	RESTORE_PCI(PCI_BASE_ADDRESS_4);
	RESTORE_PCI(PCI_INTERRUPT_LINE);
	RESTORE_PCI(PCI_CACHE_LINE_SIZE);
	RESTORE_PCI(PCI_LATENCY_TIMER);
	pci_write_config_dword(chip->pci, PCI_COMMAND, 0x07);

	for (i = 0; i < 16; i++) {
		pci_read_config_dword(chip->pci, i * 4, &config[i]);
		LOG(1, "After X-Fi mode: PCI[0x%x] = 0x%x\n", i, config[i]);
	}

}

static int sbxfi_init_hw(struct sbxfi *chip)
{
	unsigned int gctl, gpio;
	unsigned long timeout;

	LOG(1, "INIT HW\n");
	/* reset */
	gctl = sbxfi_read(chip, GLOBAL_CTL);
	sbxfi_write(chip, GLOBAL_CTL, gctl & ~GCTL_EAI);
	sbxfi_write(chip, GLOBAL_CTL, gctl | GCTL_EAI);
	sbxfi_read(chip, GLOBAL_CTL); /* post */
	msleep(1);
	timeout = jiffies + msecs_to_jiffies(1000);
	do {
		gctl = sbxfi_read(chip, GLOBAL_CTL);
		if (gctl & GCTL_AID)
			goto ok;
		udelay(10);
	} while (time_after(timeout, jiffies));
	printk(KERN_ERR "XFi reset timeout\n");
	return -EIO;

 ok:
	/* bypass transport ring, bypass filter, bypass tank, bypass DSP */
	sbxfi_write(chip, GLOBAL_CTL, gctl | GCTL_ET | GCTL_FBP | GCTL_TBP |
		    GCTL_DBP | GCTL_EAI | GCTL_EAC);
	sbxfi_read(chip, GLOBAL_CTL); /* post */
	msleep(10);

	sbxfi_write(chip, GLOBAL_INT_ENABLE, 0);
	sbxfi_write(chip, SRCIP(0), 0);
	sbxfi_read(chip, SRCIP(0)); /* post */
	msleep(30);

	switch (sbxfi_read(chip, PLL_CTL)) {
	case 0x1480a001:
	case 0x1480a731:
		break;
	default:
		sbxfi_write(chip, PLL_CTL, 0x1480a001);
		break;
	}
	sbxfi_read(chip, PLL_CTL);
	msleep(40);
	sbxfi_read(chip, PLL_CTL);

	/* set up GPIO */
	switch (chip->type) {
	case SBXFI_HW_055x:
		gpio = GPIO_O_055X_AUX_MICLINE_SEL |	/* aux */
			GPIO_O_055X_MIC_SW |		/* f/b mic */
			GPIO_O_055X_DIGITAL_IO_SEL |	/* analog */
			GPIO_O_055X_MIC_POWER_ON |	/* mic power-on */
			GPIO_O_055X_LINEOUT1_MUTE |	/* lineout1 unmute */
			GPIO_O_055X_EXDRV_DAC_96K |	/* extdrv DAC 96kHz */
			GPIO_O_055X_FP_MIC_SELECT |	/* FP mic */
			GPIO_O_055X_SPDO0_MUTE |	/* SPDIF-out0 unmute */
			GPIO_O_055X_ADC_BOOST |		/* ADC boost on */
			GPIO_O_055X_ANALOG_IN_MUTE;	/* analog-in unmute */
		break;
	case SBXFI_HW_073x:
		gpio = GPIO_O_073X_RESET_DAC |
			GPIO_O_073X_MIC_LINE_SELECT |	/* mic-in */
			GPIO_O_073X_BACK_LINEOUT_MUTE |	/* unmute line-outs */
			GPIO_O_073X_INDRV_DAC_96K |	/* intdrv DAC 96kHz */
			GPIO_O_073X_INDRV_SPDIF_BYPASS;	/* no bypass */
		break;
	case SBXFI_HW_UAA:
		gpio = GPIO_O_UAA_RESET_DAC |
			GPIO_O_UAA_MIC_LINE_SELECT |	/* mic-in */
			GPIO_O_UAA_BACK_LINEOUT_MUTE;	/* unmute line-outs */
		break;
	default:
		gpio = GPIO_O_RESET_DAC |
			GPIO_O_DIGITAL_IO_SELECT |	/* analog */
			GPIO_O_MIC_LINE_SELECT |	/* mic */
			GPIO_O_BACK_LINEOUT_MUTE |	/* unmute line-outs */
			GPIO_O_INDRV_DAC_96K |		/* intdrv DAC 96kHz */
			GPIO_O_INDRV_SPDIF_BYPASS;	/* no bypass */
		break;
	}
	sbxfi_write(chip, GPIO_CTL, gpio);

	sbxfi_write(chip, TRANS_CTL, 0x13);
	sbxfi_write(chip, TRANS_INT_STAT, 0x200c01);

	switch (base_rate) {
	case 96000:
	case 48000:
		break;
	default:
		printk(KERN_ERR PFX "Invalid base_rate %d:"
		       "must be either 48000 or 96000\n", base_rate);
		base_rate = 96000;
		break;
	}
	chip->base_rate = base_rate;

	sbxfi_daoimap_clear(chip);

	sbxfi_init_dac(chip);
	sbxfi_setup_i2s(chip);

	return 0;
}

static int __devinit sbxfi_create(struct snd_card *card, struct pci_dev *pci,
				  struct sbxfi **rchip)
{
	struct sbxfi *chip;
	int err;
	static struct snd_device_ops ops = {
		.dev_free = sbxfi_dev_free,
	};

	*rchip = NULL;

	err = pci_enable_device(pci);
	if (err < 0)
		return err;

	chip = kzalloc(sizeof(*chip), GFP_KERNEL);
	if (!chip)
		return -ENOMEM;
	chip->card = card;
	chip->pci = pci;
	chip->irq = -1;
	spin_lock_init(&chip->reg_lock);
	spin_lock_init(&chip->port_lock);
	spin_lock_init(&chip->timer_lock);
	mutex_init(&chip->tlb_mutex);
	INIT_LIST_HEAD(&chip->port_list);
	sbxfi_init_timer(chip);

	strcpy(card->driver, "SBXFi");

	if ((pci->subsystem_device & 0x6000) == 0x6000)
		chip->type = SBXFI_HW_UAA;
	else {
		switch (pci->subsystem_device) {
		case 0x0029:
		case 0x0031:
			chip->type = SBXFI_HW_073x;
			break;
		case 0x0022:
		case 0x002f:
			chip->type = SBXFI_HW_055x;
			break;
		default:
			chip->type = SBXFI_HW_ORIG;
			break;
		}
	}

	pci_set_master(pci);

	/* VISTA-compatible hardware has initially memory access in bar0;
	 * it will be switched to I/O access after XFi mode
	 */
	if (chip->type == SBXFI_HW_UAA)
		chip->bar = 5;
	else /* native XFi mode */
		chip->bar = 0;
	err = pci_request_region(pci, chip->bar, "SB-XFi");
	if (err < 0) {
		printk(KERN_ERR PFX "unable to grab PCI bar %d\n", chip->bar);
		err = -EBUSY;
		goto error;
	}
	chip->port = pci_resource_start(chip->pci, chip->bar);

	if (chip->type == SBXFI_HW_UAA)
		sbxfi_switch_xfi_mode(chip);

	if (request_irq(pci->irq, sbxfi_interrupt, IRQF_SHARED,
			"SB-XFi", chip)) {
		printk(KERN_ERR PFX "unable to grab IRQ %d\n", pci->irq);
		err = -EBUSY;
		goto error;
	}
	chip->irq = pci->irq;

	err = sbxfi_init_hw(chip);
	if (err < 0)
		goto error;

	if (pci_set_dma_mask(pci, SBXFI_DMA_MASK) < 0 ||
	    pci_set_consistent_dma_mask(pci, SBXFI_DMA_MASK) < 0) {
		printk(KERN_ERR PFX "Unable to set DMA mask\n");
		err = -EINVAL;
		goto error;
	}

	err = sbxfi_alloc_buffer(chip);
	if (err < 0)
		goto error;

	err = snd_device_new(card, SNDRV_DEV_LOWLEVEL, chip, &ops);
	if (err <0) {
		snd_printk(KERN_ERR PFX "Error creating device [card]!\n");
		goto error;
	}

	sprintf(card->shortname, "SB XFi (%s)", typenames[chip->type]);
	sprintf(card->longname, "%s at 0x%lx irq %i", card->shortname,
		chip->port, chip->irq);

	*rchip = chip;
	return 0;

 error:
	sbxfi_free(chip);
	return err;
}

static int __devinit sbxfi_probe(struct pci_dev *pci,
				 const struct pci_device_id *pci_id)
{
	static int dev;
	struct snd_card *card;
	struct sbxfi *chip;
	int err;

	if (dev >= SNDRV_CARDS)
		return -ENODEV;
	if (!enable[dev]) {
		dev++;
		return -ENOENT;
	}

	card = snd_card_new(index[dev], id[dev], THIS_MODULE, 0);
	if (!card) {
		snd_printk(KERN_ERR PFX "Error creating card!\n");
		return -ENOMEM;
	}

	err = sbxfi_create(card, pci, &chip);
	if (err < 0)
		goto error;

	err = sbxfi_create_pcm(chip);
	if (err < 0)
		goto error;

	err = sbxfi_create_mixer(chip);
	if (err < 0)
		goto error;

	sbxfi_proc_init(chip);

	snd_card_set_dev(card, &pci->dev);
	err = snd_card_register(card);
	if (err < 0)
		goto error;

	pci_set_drvdata(pci, card);

	dev++;
	return 0;

 error:
	snd_card_free(card);
	return err;
}

static void __devexit sbxfi_remove(struct pci_dev *pci)
{
	snd_card_free(pci_get_drvdata(pci));
	pci_set_drvdata(pci, NULL);
}


/* PCI IDs */
static struct pci_device_id sbxfi_ids[] = {
	{ PCI_DEVICE(0x1102, 0x0005) }, /* SB XFi */
	{ PCI_DEVICE(0x1102, 0x0009) }, /* SB XFi-E */
	{ PCI_DEVICE(0x1102, 0x000b) }, /* SB XFi-E2 */
	{ 0, }
};
MODULE_DEVICE_TABLE(pci, sbxfi_ids);

static struct pci_driver driver = {
	.name = "SB-XFi",
	.id_table = sbxfi_ids,
	.probe = sbxfi_probe,
	.remove = __devexit_p(sbxfi_remove),
	/* FIXME: suspend/resume */
};

static int __init alsa_card_sbxfi_init(void)
{
	return pci_register_driver(&driver);
}

static void __exit alsa_card_sbxfi_exit(void)
{
	pci_unregister_driver(&driver);
}

module_init(alsa_card_sbxfi_init)
module_exit(alsa_card_sbxfi_exit)
